# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014-2020)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.astro.range`
"""

from unittest import mock

import pytest

from astropy import units
from scipy.integrate import trapezoid

from ... import astro
from ...testing import utils
from ...timeseries import TimeSeries
from ...frequencyseries import FrequencySeries
from ...spectrogram import Spectrogram

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__credits__ = 'Alex Urban <alexander.urban@ligo.org>'

TEST_RESULTS = {
    'sensemon_range': 19.332958991178117 * units.Mpc,
    'inspiral_range': 18.519899937121536 * units.Mpc,
    'inspiral_range_snr_10': 14.82352747 * units.Mpc,
    'burst_range': 13.592140825568954 * units.Mpc,
}


@pytest.fixture(scope='module')
def hoft():
    return TimeSeries.read(
        utils.TEST_HDF5_FILE,
        "L1:LDAS-STRAIN",
        format="hdf5",
    )


@pytest.fixture(scope='module')
def psd(hoft):
    return hoft.psd(
        .4,
        overlap=.2,
        method="median",
        window=('kaiser', 24),
    )


# -- sensemon -----------------------------------

def test_sensemon_range_psd(psd):
    """Test for :func:`gwpy.astro.sensemon_range_psd`
    """
    fisco = astro.range._get_isco_frequency(1.4, 1.4).value
    frange = (psd.frequencies.value < fisco)
    r = astro.sensemon_range_psd(psd[frange])
    assert isinstance(r, FrequencySeries)
    utils.assert_quantity_almost_equal(
        trapezoid(r, r.frequencies) ** (1/2.),
        TEST_RESULTS['sensemon_range'],
    )
    assert r.f0.value > 0


def test_sensemon_range(psd):
    """Test for :func:`gwpy.astro.sensemon_range`
    """
    r = astro.sensemon_range(psd)
    utils.assert_quantity_almost_equal(r, TEST_RESULTS['sensemon_range'])


# -- inspiral-range -----------------------------


@mock.patch.dict("sys.modules", {"inspiral_range": None})
def test_inspiral_range_missing_dep(psd):
    with pytest.raises(
        ModuleNotFoundError,
        match="extra package 'inspiral-range'",
    ):
        astro.inspiral_range(psd)


@pytest.mark.requires("inspiral_range")
def test_inspiral_range_psd(psd):
    """Test for :func:`gwpy.astro.inspiral_range_psd`
    """
    frange = (psd.frequencies.value < 4096)
    r = astro.inspiral_range_psd(psd[frange])
    assert isinstance(r, FrequencySeries)
    utils.assert_quantity_almost_equal(
        trapezoid(r, r.frequencies) ** (1/2.),
        TEST_RESULTS['inspiral_range'],
    )
    assert r.f0.value > 0


@pytest.mark.requires("inspiral_range")
@pytest.mark.parametrize(("snr", "expected"), (
    (8, TEST_RESULTS["inspiral_range"]),
    (10, TEST_RESULTS["inspiral_range_snr_10"]),
))
def test_inspiral_range(psd, snr, expected):
    """Test for :func:`gwpy.astro.inspiral_range`
    """
    result = astro.inspiral_range(psd, snr=snr)
    utils.assert_quantity_almost_equal(result, expected)


# -- burst range --------------------------------

def test_burst_range_spectrum(psd):
    """Test for :func:`gwpy.astro.burst_range_spectrum`
    """
    f = psd.frequencies
    frange = (f.value >= 100) & (f.value < 500)
    r = astro.burst_range_spectrum(psd[frange])
    assert isinstance(r, FrequencySeries)
    utils.assert_quantity_almost_equal(
        (trapezoid(r**3, f[frange]) / (400 * units.Hz)) ** (1/3.),
        TEST_RESULTS['burst_range'],
    )
    assert r.f0.value > 0


def test_burst_range(psd):
    """Test for :func:`gwpy.astro.burst_range`
    """
    r = astro.burst_range(psd)
    utils.assert_quantity_almost_equal(r, TEST_RESULTS['burst_range'])


# -- timeseries/spectrogram wrappers ------------

@pytest.mark.parametrize('rangekwargs', [
    pytest.param(
        {'mass1': 1.4, 'mass2': 1.4},
        marks=[pytest.mark.requires("inspiral_range")],
        id="inspiral_range",
    ),
    pytest.param(
        {'mass1': 1.4, 'mass2': 1.4, "range_func": astro.sensemon_range},
        id="sensemon_range",
    ),
    pytest.param({'energy': 1e-2}, id="burst_range"),
])
def test_range_timeseries(hoft, rangekwargs):
    trends = astro.range_timeseries(
        hoft,
        0.5,
        fftlength=0.25,
        overlap=0.125,
        method="median",
        nproc=2,
        **rangekwargs,
    )
    assert isinstance(trends, TimeSeries)
    assert trends.size == 2
    assert trends.unit == 'Mpc'
    assert trends.dt == 0.5 * units.second


@pytest.mark.parametrize('rangekwargs, outunit', [
    pytest.param(
        {'mass1': 1.4, 'mass2': 1.4},
        units.Mpc ** 2 / units.Hz,
        marks=[pytest.mark.requires("inspiral_range")],
        id="inspiral_range",
    ),
    pytest.param(
        {'mass1': 1.4, 'mass2': 1.4, "range_func": astro.sensemon_range_psd},
        units.Mpc ** 2 / units.Hz,
        id="sensemon_range",
    ),
    pytest.param({'energy': 1e-2}, units.Mpc, id="burst_range"),
])
def test_range_spectrogram(hoft, rangekwargs, outunit):
    spec = astro.range_spectrogram(
        hoft,
        0.5,
        fftlength=0.25,
        overlap=0.125,
        method="median",
        nproc=2,
        **rangekwargs,
    )
    assert isinstance(spec, Spectrogram)
    assert spec.shape[0] == 2
    assert spec.unit == outunit
    assert spec.f0 == spec.df
    assert spec.dt == 0.5 * units.second
    assert spec.df == 4 * units.Hertz


@pytest.mark.parametrize('range_func', [
    astro.range_timeseries,
    astro.range_spectrogram,
])
def test_range_incompatible_input(range_func):
    with pytest.raises(
        TypeError,
        match="Could not produce a spectrogram from the input",
    ):
        range_func(2, 0.5)



# -*- coding: utf-8 -*-
# Copyright (C) Cardiff University (2019-2022)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.table.io.pycbc` and its integration with `EventTable`.
"""

import pytest

import numpy
from numpy.random import randn

import h5py

from .. import EventTable
from ..filter import filter_table
from ..io import pycbc as io_pycbc
from ...frequencyseries import FrequencySeries
from ...testing.utils import (
    assert_array_equal,
    assert_quantity_sub_equal,
    assert_table_equal,
)


# -- fixtures -----------------------------------

@pytest.fixture
def h5file():
    """Create an empty in-memory HDF5 file.
    """
    with h5py.File(
        'test',
        mode='w-',
        driver='core',
        backing_store=False,
    ) as h5f:
        yield h5f


@pytest.fixture
def pycbclivetable():
    """A populated `EventTable` in PyCBC format.
    """
    names = [
        'a',
        'b',
        'c',
        'chisq',
        'd',
        'e',
        'f',
        'mass1',
        'mass2',
        'snr',
    ]
    rows = []
    for i, name in enumerate(names):
        rows.append(randn(100) * 1000)
    return EventTable(rows, names=names)


@pytest.fixture
def pycbclivepsd():
    """A mock PSD.
    """
    return FrequencySeries(randn(1000), df=1)


@pytest.fixture
def pycbclivefile(tmp_path, pycbclivetable, pycbclivepsd):
    """A fully-formed PyCBC-format HDF5 file.
    """
    # create table
    loudest = (pycbclivetable['snr'] > 500).nonzero()[0]

    # manually create pycbc_live-format HDF5 file
    tmp = tmp_path / "X1-Live-0-0.hdf"
    with h5py.File(tmp, "w") as h5f:
        group = h5f.create_group('X1')
        for col in pycbclivetable.columns:
            group.create_dataset(data=pycbclivetable[col], name=col)
        group.create_dataset('loudest', data=loudest)
        group.create_dataset('psd', data=pycbclivepsd.value)
        group['psd'].attrs['delta_f'] = pycbclivepsd.df.to('Hz').value

    return tmp


# -- internal tests -----------------------------


@pytest.mark.parametrize(("filename", "result"), [
    ("X1-Live-0-0.h5", True),
    ("X1-Live-0-0.hdf5", True),
    ("X1-Live-0-0.h5", True),
    ("X1-MY_DATA-0-0.h5", False),
])
def test_idenfity_pycbc_live(tmp_path, filename, result):
    path = str(tmp_path / filename)
    h5py.File(path, "w").close()
    assert io_pycbc.identify_pycbc_live("read", path, None) is result
    with open(path, "rb") as h5f:  # check with open file as well
        assert io_pycbc.identify_pycbc_live("read", path, h5f) is result


def test_empty_hdf5_file(h5file):
    """Check that :func:`empty_hdf5_file` works.
    """
    assert io_pycbc.empty_hdf5_file(h5file)


def test_empty_hdf5_file_group(h5file):
    """Check that :func:`empty_hdf5_file` works with (empty) groups.
    """
    h5file.create_group("H1")
    assert io_pycbc.empty_hdf5_file(h5file)
    assert io_pycbc.empty_hdf5_file(h5file, ifo="H1")


def test_empty_hdf5_file_datasets(h5file):
    """Check that :func:`empty_hdf5_file` works with (empty) datasets.
    """
    h1group = h5file.create_group("H1")
    h1group.create_dataset("psd", data=numpy.empty(10))
    assert io_pycbc.empty_hdf5_file(h5file, ifo="H1")

    h1group.create_dataset("gates", data=numpy.empty(10))
    assert io_pycbc.empty_hdf5_file(h5file, ifo="H1")

    h1group.create_dataset("time", data=numpy.empty(10))
    assert not io_pycbc.empty_hdf5_file(h5file, ifo="H1")


# -- EventTable integration tests ---------------

@pytest.mark.parametrize("fmt", [
    None,  # should default to hdf5.pycbc_live as the highest-priority format
    "hdf5.pycbc_live",
])
def test_read_pycbc_live(pycbclivetable, pycbclivefile, fmt):
    """Check that `EventTable` can read a PyCBC-Live file.
    """
    table = EventTable.read(pycbclivefile, format=fmt)
    assert_table_equal(pycbclivetable, table)
    assert table.meta['ifo'] == 'X1'


def test_read_pycbc_live_kwargs(pycbclivetable, pycbclivefile):
    """Check that `EventTable` can read a PyCBC-Live file using keywords.
    """
    table = EventTable.read(
        pycbclivefile,
        format='hdf5.pycbc_live',
        ifo='X1',
    )
    assert_table_equal(pycbclivetable, table)


def test_read_pycbc_live_loudest(pycbclivetable, pycbclivefile):
    """Check that `EventTable` can read the `loudest` table from
    a PyCBC-Live file.
    """
    table = EventTable.read(
        pycbclivefile,
        format="hdf5.pycbc_live",
        loudest=True,
    )
    assert_table_equal(pycbclivetable.filter('snr > 500'), table)


def test_read_pycbc_live_extended_metadata(
    pycbclivetable,
    pycbclivepsd,
    pycbclivefile,
):
    """Check that `EventTable` can read extended metadata from a PyCBC file.
    """
    table = EventTable.read(
        pycbclivefile,
        format="hdf5.pycbc_live",
        extended_metadata=True,  # default
    )
    assert_table_equal(pycbclivetable, table)
    assert_array_equal(
        table.meta['loudest'],
        (pycbclivetable['snr'] > 500).nonzero()[0],
    )
    assert_quantity_sub_equal(
        table.meta['psd'],
        pycbclivepsd,
        exclude=['name', 'channel', 'unit', 'epoch'])


def test_read_pycbc_live_extended_metadata_false(
    pycbclivetable,
    pycbclivefile,
):
    """Check that `EventTable` can read a PyCBC file without extended metadata.
    """
    # check extended_metadata=False works
    table = EventTable.read(
        pycbclivefile,
        format="hdf5.pycbc_live",
        extended_metadata=False,
    )
    assert table.meta == {'ifo': 'X1'}


def test_read_pycbc_live_multiple_ifos(
    pycbclivetable,
    pycbclivefile,
):
    """Check that `EventTable` can handle multiple IFOs in a PyCBC-Live file
    """
    with h5py.File(pycbclivefile, "r+") as h5f:
        h5f.create_group('Z1')
    with pytest.raises(
        ValueError,
        match="PyCBC live HDF5 file contains dataset groups",
    ):
        EventTable.read(pycbclivefile, format="hdf5.pycbc_live")

    # but check that we can still read the original
    table = EventTable.read(
        pycbclivefile,
        format='hdf5.pycbc_live',
        ifo='X1',
    )
    assert_table_equal(pycbclivetable, table)


def test_read_pycbc_live_processed_columns(
    pycbclivetable,
    pycbclivefile,
):
    """Check that `EventTable` can read processed columns from a PyCBC file.
    """
    # assert processed colums works
    table = EventTable.read(
        pycbclivefile,
        format="hdf5.pycbc_live",
        ifo="X1",
        columns=["mchirp", "new_snr"],
    )
    mchirp = (
        (pycbclivetable['mass1'] * pycbclivetable['mass2']) ** (3/5.)
        / (pycbclivetable['mass1'] + pycbclivetable['mass2']) ** (1/5.)
    )
    assert_array_equal(table['mchirp'], mchirp)


def test_read_pycbc_live_selection_columns(
    pycbclivetable,
    pycbclivefile,
):
    """Check that the selection and columns kwargs work when
    reading from a PyCBC-Live file.
    """
    # test with selection and columns
    table = EventTable.read(
        pycbclivefile,
        format='hdf5.pycbc_live',
        ifo='X1',
        selection='snr>.5',
        columns=("a", "b", "mass1"),
    )
    assert_table_equal(
        table,
        filter_table(pycbclivetable, 'snr>.5')[("a", "b", "mass1")],
    )


def test_read_pycbc_live_regression_1081(
    pycbclivetable,
    pycbclivefile,
):
    """Check against regression of gwpy/gwpy#1081.
    """
    table = EventTable.read(
        pycbclivefile,
        format='hdf5.pycbc_live',
        ifo='X1',
        selection='snr>.5',
        columns=("a", "b", "snr"),
    )
    assert_table_equal(
        table,
        filter_table(pycbclivetable, 'snr>.5')[("a", "b", "snr")],
    )



# -*- coding: utf-8 -*-
# Copyright (C) Cardiff University (2021)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Custom pytest marks for GWpy

This is mainly to allow downstream users/builders/packagers to skip tests
that require annoying/flaky configuration, or just take too long, e.g.
tests that use CVMFS could be skipped via

    $ python -m pytest --pyargs gwpy -m 'not cvmfs'

This module is imported in :mod:`gwpy.conftest` such that all marks are
registered up front and visible via ``python -m pytest gwpy --markers``.
"""

MARKS = {
    # nothing here, used to support 'cvmfs' but that was removed
}


def _register_marks(config):
    """Register all marks for GWpy

    This function is designed to be called from :mod:`gwpy.conftest`.
    """
    for name, doc in MARKS.items():
        config.addinivalue_line(
            "markers",
            "{}: {}".format(name, doc),
        )



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014-2020)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Basic utilities for reading/writing LIGO_LW-format XML files.

All specific unified input/output for class objects should be placed in
an 'io' subdirectory of the containing directory for that class.
"""

import os
import os.path
from contextlib import contextmanager
from functools import wraps
from importlib import import_module

import numpy

try:
    from ligo.lw.ligolw import (
        ElementError as LigolwElementError,
        LIGOLWContentHandler,
    )
except ImportError:  # no ligo.lw
    LigolwElementError = None
    LIGOLWContentHandler = None

from .utils import (file_list, FILE_LIKE)
from ..utils.decorators import deprecated_function

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

# XML elements
XML_SIGNATURE = b'<?xml'
LIGOLW_SIGNATURE = b'<!doctype ligo_lw'
LIGOLW_ELEMENT = b'<ligo_lw>'


# -- hack around around TypeError from LIGOTimeGPS(numpy.int32(...)) ----------

def _ligotimegps(s, ns=0):
    """Catch TypeError and cast `s` and `ns` to `int`
    """
    from lal import LIGOTimeGPS
    try:
        return LIGOTimeGPS(s, ns)
    except TypeError:
        return LIGOTimeGPS(int(s), int(ns))


@contextmanager
def patch_ligotimegps(module="ligo.lw.lsctables"):
    """Context manager to on-the-fly patch LIGOTimeGPS to accept all int types
    """
    module = import_module(module)
    orig = module.LIGOTimeGPS
    module.LIGOTimeGPS = _ligotimegps
    try:
        yield
    finally:
        module.LIGOTimeGPS = orig


# -- content handling ---------------------------------------------------------

def _int_ilwd(ilwd):
    """Convert an ``ilwd`` string into an integer.
    """
    try:
        _, _, i = ilwd.strip().split(":")
    except ValueError:
        raise ValueError(f"invalid ilwd:char '{ilwd}'")
    return int(i)


def strip_ilwdchar(_ContentHandler):
    """Wrap a LIGO_LW content handler to swap ilwdchar for int on-the-fly
    when reading a document

    This is adapted from :func:`ligo.skymap.utils.ilwd`, copyright
    Leo Singer (GPL-3.0-or-later).
    """
    from ligo.lw.lsctables import TableByName
    from ligo.lw.table import (Column, TableStream)
    from ligo.lw.types import FromPyType

    class IlwdMapContentHandler(_ContentHandler):

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._idconverter = {}

        @wraps(_ContentHandler.startColumn)
        def startColumn(self, parent, attrs):
            result = super().startColumn(parent, attrs)

            # if an old ID type, convert type definition to an int
            if result.Type == "ilwd:char":
                self._idconverter[(id(parent), result.Name)] = _int_ilwd
                result.Type = FromPyType[int]

            try:
                validcolumns = TableByName[parent.Name].validcolumns
            except KeyError:  # parent.Name not in TableByName
                return result
            if result.Name not in validcolumns:
                stripped_column_to_valid_column = {
                    Column.ColumnName(name): name
                    for name in validcolumns
                }
                if result.Name in stripped_column_to_valid_column:
                    result.setAttribute(
                        'Name',
                        stripped_column_to_valid_column[result.Name],
                    )

            return result

        @wraps(_ContentHandler.startStream)
        def startStream(self, parent, attrs):
            result = super().startStream(parent, attrs)
            if isinstance(result, TableStream):
                loadcolumns = set(parent.columnnames)
                if parent.loadcolumns is not None:
                    loadcolumns &= set(parent.loadcolumns)
                pid = id(parent)
                result._tokenizer.set_types([
                    self._idconverter.pop((pid, colname), pytype)
                    if colname in loadcolumns else None
                    for pytype, colname in zip(
                        parent.columnpytypes,
                        parent.columnnames,
                    )
                ])
            return result

    return IlwdMapContentHandler


def _wrap_content_handler(contenthandler):
    from ligo.lw.lsctables import use_in

    @strip_ilwdchar
    @use_in
    class ContentHandler(contenthandler):
        pass

    return ContentHandler


def default_content_handler():
    """Return a standard content handler to read LIGO_LW documents

    This handler knows how to parse LSCTables, and automatically converts
    old-style ilwdchar ID types to `int`.

    Returns
    -------
    contenthandler : subclass of `ligo.lw.ligolw.LIGOLWContentHandler`
    """
    from ligo.lw.ligolw import LIGOLWContentHandler
    return _wrap_content_handler(LIGOLWContentHandler)


def get_partial_contenthandler(element):
    """Build a `PartialLIGOLWContentHandler` to read only this element

    Parameters
    ----------
    element : `type`, subclass of :class:`~ligo.lw.ligolw.Element`
        the element class to be read

    Returns
    -------
    contenthandler : `type`
        a subclass of `~ligo.lw.ligolw.PartialLIGOLWContentHandler`
        to read only the given `element`
    """
    from ligo.lw.ligolw import PartialLIGOLWContentHandler
    from ligo.lw.table import Table

    if issubclass(element, Table):
        def _element_filter(name, attrs):
            return element.CheckProperties(name, attrs)
    else:
        def _element_filter(name, _):
            return name == element.tagName

    return build_content_handler(PartialLIGOLWContentHandler, _element_filter)


def get_filtering_contenthandler(element):
    """Build a `FilteringLIGOLWContentHandler` to exclude this element

    Parameters
    ----------
    element : `type`, subclass of :class:`~ligo.lw.ligolw.Element`
        the element to exclude (and its children)

    Returns
    -------
    contenthandler : `type`
        a subclass of `~ligo.lw.ligolw.FilteringLIGOLWContentHandler`
        to exclude an element and its children
    """
    from ligo.lw.ligolw import FilteringLIGOLWContentHandler
    from ligo.lw.table import Table

    if issubclass(element, Table):
        def _element_filter(name, attrs):
            return ~element.CheckProperties(name, attrs)
    else:
        def _element_filter(name, _):
            # pylint: disable=unused-argument
            return name != element.tagName

    return build_content_handler(
        FilteringLIGOLWContentHandler,
        _element_filter,
    )


def build_content_handler(parent, filter_func):
    """Build a `~xml.sax.handler.ContentHandler` with a given filter

    Parameters
    ----------
    parent : `type`, subclass of `xml.sax.handler.ContentHandler`
        a class of contenthandler to use

    filter_func : `callable`
        the filter function to pass to the content handler creation

    Returns
    -------
    contenthandler : subclass of ``parent``
        a new content handler that applies the filter function and the
        default parsing extras from :func:`_wrap_content_handler`.
    """
    class ContentHandler(parent):
        # pylint: disable=too-few-public-methods
        def __init__(self, document):
            super().__init__(document, filter_func)

    return _wrap_content_handler(ContentHandler)


# -- reading ------------------------------------------------------------------

def read_ligolw(source, contenthandler=None, **kwargs):
    """Read one or more LIGO_LW format files

    Parameters
    ----------
    source : `str`, `file`
        the open file or file path to read

    contenthandler : `~xml.sax.handler.ContentHandler`, optional
        content handler used to parse document

    verbose : `bool`, optional
        be verbose when reading files, default: `False`

    Returns
    -------
    xmldoc : :class:`~ligo.lw.ligolw.Document`
        the document object as parsed from the file(s)
    """
    from ligo.lw.ligolw import Document
    from ligo.lw import types
    from ligo.lw.utils import (load_url, ligolw_add)

    # mock ToPyType to link to numpy dtypes
    topytype = types.ToPyType.copy()
    for key in types.ToPyType:
        if key in types.ToNumPyType:
            types.ToPyType[key] = numpy.dtype(types.ToNumPyType[key]).type

    # set contenthandler
    if contenthandler is None:
        contenthandler = default_content_handler()

    # read one or more files into a single Document
    source = file_list(source)
    try:
        if len(source) == 1:
            return load_url(
                source[0],
                contenthandler=contenthandler,
                **kwargs
            )
        return ligolw_add.ligolw_add(
            Document(),
            source,
            contenthandler=contenthandler,
            **kwargs
        )
    finally:  # replace ToPyType
        types.ToPyType = topytype


# -- reading ------------------------------------------------------------------

def read_table(
    source,
    tablename=None,
    columns=None,
    contenthandler=None,
    **kwargs,
):
    """Read a :class:`~ligo.lw.table.Table` from one or more LIGO_LW files

    Parameters
    ----------
    source : `Document`, `file`, `str`, `CacheEntry`, `list`
        object representing one or more files. One of

        - a LIGO_LW :class:`~ligo.lw.ligolw.Document`
        - an open `file`
        - a `str` pointing to a file path on disk
        - a formatted :class:`~lal.utils.CacheEntry` representing one file
        - a `list` of `str` file paths or :class:`~lal.utils.CacheEntry`

    tablename : `str`
        name of the table to read.

    columns : `list`, optional
        list of column name strings to read, default all.

    contenthandler : `~xml.sax.handler.ContentHandler`, optional
        SAX content handler for parsing LIGO_LW documents.

    **kwargs
        other keyword arguments are passed to `~gwpy.io.ligolw.read_ligolw`

    Returns
    -------
    table : :class:`~ligo.lw.table.Table`
        `Table` of data
    """
    from ligo.lw.ligolw import Document
    from ligo.lw import (table, lsctables)

    # get content handler to read only this table (if given)
    if tablename is not None:
        tableclass = lsctables.TableByName[
            table.Table.TableName(tablename)
        ]
        if contenthandler is None:
            contenthandler = get_partial_contenthandler(tableclass)

        # overwrite loading column names to get just what was asked for
        _oldcols = tableclass.loadcolumns
        if columns is not None:
            tableclass.loadcolumns = columns

    # read document
    if isinstance(source, Document):
        xmldoc = source
    else:
        if contenthandler is None:
            contenthandler = default_content_handler()
        try:
            xmldoc = read_ligolw(
                source,
                contenthandler=contenthandler,
                **kwargs)
        finally:  # reinstate original set of loading column names
            if tablename is not None:
                tableclass.loadcolumns = _oldcols

    # now find the right table
    if tablename is None:
        tables = list_tables(xmldoc)
        if not tables:
            raise ValueError("No tables found in LIGO_LW document(s)")
        if len(tables) > 1:
            raise ValueError(
                "Multiple tables found in LIGO_LW document(s), please specify "
                "the table to read via the ``tablename=`` keyword argument. "
                "The following tables were found: "
                "'{}'".format("', '".join(tables)),
            )
        tableclass = lsctables.TableByName[table.Table.TableName(tables[0])]

    # extract table
    return tableclass.get_table(xmldoc)


# -- writing ------------------------------------------------------------------

def open_xmldoc(fobj, contenthandler=None, **kwargs):
    """Try and open an existing LIGO_LW-format file, or create a new Document

    Parameters
    ----------
    fobj : `str`, `file`
        file path or open file object to read

    contenthandler : `~xml.sax.handler.ContentHandler`, optional
        the content handler with which to parse the document, if not given
        a default handler will be created using
        :func:`default_content_handler`.

    **kwargs
        other keyword arguments to pass to
        :func:`~ligo.lw.utils.load_fileobj` as appropriate

    Returns
    --------
    xmldoc : :class:`~ligo.lw.ligolw.Document`
        either the `Document` as parsed from an existing file, or a new, empty
        `Document`
    """
    from ligo.lw.ligolw import Document
    from ligo.lw.utils import load_fileobj

    if contenthandler is None:
        contenthandler = default_content_handler()

    # read from an existing Path/filename
    if not isinstance(fobj, FILE_LIKE):
        try:
            with open(fobj, "rb") as fobj2:
                return open_xmldoc(
                    fobj2,
                    contenthandler=contenthandler,
                    **kwargs,
                )
        except (OSError, IOError):
            # or just create a new Document
            return Document()

    return load_fileobj(
        fobj,
        contenthandler=contenthandler,
        **kwargs,
    )


def get_ligolw_element(xmldoc):
    """Find an existing <LIGO_LW> element in this XML Document
    """
    from ligo.lw.ligolw import (LIGO_LW, WalkChildren)

    if isinstance(xmldoc, LIGO_LW):
        return xmldoc
    for elem in WalkChildren(xmldoc):
        if isinstance(elem, LIGO_LW):
            return elem
    raise ValueError("Cannot find LIGO_LW element in XML Document")


def write_tables_to_document(xmldoc, tables, overwrite=False):
    """Write the given LIGO_LW table into a :class:`Document`

    Parameters
    ----------
    xmldoc : :class:`~ligo.lw.ligolw.Document`
        the document to write into

    tables : `list` of :class:`~ligo.lw.table.Table`
        the set of tables to write

    overwrite : `bool`, optional, default: `False`
        if `True`, delete an existing instance of the table type, otherwise
        append new rows
    """
    from ligo.lw.ligolw import LIGO_LW
    from ligo.lw import lsctables

    # find or create LIGO_LW tag
    try:
        llw = get_ligolw_element(xmldoc)
    except ValueError:
        llw = LIGO_LW()
        xmldoc.appendChild(llw)

    for table in tables:
        try:  # append new data to existing table
            old = lsctables.TableByName[
                table.TableName(table.Name)].get_table(xmldoc)
        except ValueError:  # or create a new table
            llw.appendChild(table)
        else:
            if overwrite:
                llw.removeChild(old)
                old.unlink()
                llw.appendChild(table)
            else:
                old.extend(table)

    return xmldoc


def write_tables(
    target,
    tables,
    append=False,
    overwrite=False,
    contenthandler=None,
    **kwargs,
):
    """Write an LIGO_LW table to file

    Parameters
    ----------
    target : `str`, `file`, :class:`~ligo.lw.ligolw.Document`
        the file or document to write into

    tables : `list`, `tuple` of :class:`~ligo.lw.table.Table`
        the tables to write

    append : `bool`, optional, default: `False`
        if `True`, append to an existing file/table, otherwise `overwrite`

    overwrite : `bool`, optional, default: `False`
        if `True`, delete an existing instance of the table type, otherwise
        append new rows

    contenthandler : `~xml.sax.handler.ContentHandler`, optional
        the content handler with which to parse the document, if not given
        a default handler will be created using
        :func:`default_content_handler`.

    **kwargs
        other keyword arguments to pass to
        :func:`~ligo.lw.utils.load_fileobj` as appropriate
    """
    from ligo.lw.ligolw import Document, LIGO_LW
    from ligo.lw import utils as ligolw_utils

    # allow writing directly to XML
    if isinstance(target, (Document, LIGO_LW)):
        xmldoc = target
    # open existing document, if possible
    elif append:
        if contenthandler is None:
            contenthandler = default_content_handler()
        xmldoc = open_xmldoc(
            target,
            contenthandler=contenthandler,
        )
    # fail on existing document and not overwriting
    elif (
        not overwrite
        and isinstance(target, (str, os.PathLike))
        and os.path.exists(target)
    ):
        raise IOError(f"File exists: {target}")
    else:  # or create a new document
        xmldoc = Document()

    # convert table to format
    write_tables_to_document(xmldoc, tables, overwrite=overwrite)

    # find writer function and target filename
    if isinstance(target, FILE_LIKE):
        writer = ligolw_utils.write_fileobj
        name = target.name
    else:
        writer = ligolw_utils.write_filename
        name = target = str(target)

    # handle gzip compression kwargs
    if name.endswith('.gz'):
        kwargs.setdefault('compress', 'gz')

    # write XML
    writer(xmldoc, target, **kwargs)


# -- utilities ----------------------------------------------------------------

def iter_tables(source):
    """Iterate over all tables in the given document(s)

    Parameters
    ----------
    source : `file`, `str`, :class:`~ligo.lw.ligolw.Document`, `list`
        one or more open files, file paths, or LIGO_LW `Document`s

    Yields
    ------
    ligo.lw.table.Table
        a table structure from the document(s)
    """
    from ligo.lw.ligolw import (Element, Stream, WalkChildren)

    # get LIGO_LW object
    if not isinstance(source, Element):
        filt = get_filtering_contenthandler(Stream)
        source = read_ligolw(source, contenthandler=filt)
    llw = get_ligolw_element(source)

    # yield tables
    for elem in WalkChildren(llw):
        if elem.tagName == "Table":
            yield elem


def list_tables(source):
    """List the names of all tables in this file(s)

    Parameters
    ----------
    source : `file`, `str`, :class:`~ligo.lw.ligolw.Document`, `list`
        one or more open files, file paths, or LIGO_LW `Document`s

    Examples
    --------
    >>> from gwpy.io.ligolw import list_tables
    >>> print(list_tables('H1-LDAS_STRAIN-968654552-10.xml.gz'))
    ['process', 'process_params', 'sngl_burst', 'search_summary', 'segment_definer', 'segment_summary', 'segment']
    """  # noqa: E501
    return [tbl.TableName(tbl.Name) for tbl in iter_tables(source)]


def to_table_type(val, cls, colname):
    """Cast a value to the correct type for inclusion in a LIGO_LW table

    This method returns the input unmodified if a type mapping for ``colname``
    isn't found.

    Parameters
    ----------
    val : `object`
        The input object to convert, of any type

    cls : `type`, subclass of :class:`~ligo.lw.table.Table`
        the table class to map against

    colname : `str`
        The name of the mapping column

    Returns
    -------
    obj : `object`
        The input ``val`` cast to the correct type

    Examples
    --------
    >>> from gwpy.io.ligolw import to_table_type as to_ligolw_type
    >>> from ligo.lw.lsctables import SnglBurstTable
    >>> x = to_ligolw_type(1.0, SnglBurstTable, 'central_freq'))
    >>> print(type(x), x)
    <class 'numpy.float32'> 1.0
    """
    from ligo.lw.types import (
        ToNumPyType as numpytypes,
        ToPyType as pytypes,
    )

    # if nothing to do...
    if val is None or colname not in cls.validcolumns:
        return val

    llwtype = cls.validcolumns[colname]

    # map to numpy or python types
    try:
        return numpy.sctypeDict[numpytypes[llwtype]](val)
    except KeyError:
        return pytypes[llwtype](val)


# -- identify -----------------------------------------------------------------

def is_ligolw(origin, filepath, fileobj, *args, **kwargs):
    """Identify a file object as LIGO_LW-format XML
    """
    # pylint: disable=unused-argument
    if fileobj is not None:
        loc = fileobj.tell()
        fileobj.seek(0)
        try:
            line1 = fileobj.readline().lower()
            line2 = fileobj.readline().lower()
            try:
                return (
                    line1.startswith(XML_SIGNATURE)
                    and line2.startswith((LIGOLW_SIGNATURE, LIGOLW_ELEMENT))
                )
            except TypeError:  # bytes vs str
                return (
                    line1.startswith(XML_SIGNATURE.decode('utf-8'))
                    and line2.startswith((
                        LIGOLW_SIGNATURE.decode('utf-8'),
                        LIGOLW_ELEMENT.decode('utf-8'),
                    ))
                )
        finally:
            fileobj.seek(loc)

    try:
        from ligo.lw.ligolw import Element
    except ImportError:
        return
    return len(args) > 0 and isinstance(args[0], Element)


@deprecated_function
def is_xml(origin, filepath, fileobj, *args, **kwargs):  # pragma: no cover
    """Identify a file object as XML (any format)
    """
    # pylint: disable=unused-argument
    if fileobj is not None:
        loc = fileobj.tell()
        fileobj.seek(0)
        try:
            sig = fileobj.read(5).lower()
            return sig == XML_SIGNATURE
        finally:
            fileobj.seek(loc)
    elif filepath is not None:
        return filepath.endswith(('.xml', '.xml.gz'))



# -*- coding: utf-8 -*-
# Copyright (C) Cardiff University (2021)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.utils.sphinx.ex2rst`
"""

import pytest

from ..sphinx import ex2rst

EXAMPLE = """
# Header

\"\"\"Example example

Something something something
\"\"\"

__author__ = "Duncan Macleod"
__credits__ = "Someone else"  # ignored
__currentmodule__ = "gwpy.timeseries"

# This is an example:
from gwpy.timeseries import TimeSeries

# and then we do this:
a = TimeSeries(12345)
plot = a.plot()
plot.show()

# and then something else
b = TimeSeries(67890)

# tada!
"""

EXAMPLE_OUTPUT = """
.. _gwpy-example-test_ex2rst0-example:

.. sectionauthor:: Duncan Macleod

.. currentmodule:: gwpy.timeseries

Example example
###############

Something something something

This is an example:

.. plot::
   :context: reset
   :nofigs:
   :include-source:

   from gwpy.timeseries import TimeSeries

and then we do this:

.. plot::
   :context:
   :include-source:

   a = TimeSeries(12345)
   plot = a.plot()
   plot.show()

and then something else

.. plot::
   :context: close-figs
   :nofigs:
   :include-source:

   b = TimeSeries(67890)

tada!
""".lstrip()


@pytest.fixture
def example(tmp_path):
    """Write the EXAMPLE to a python file and return that
    """
    source = tmp_path / "example.py"
    source.write_text(EXAMPLE)
    return source


def test_ex2rst(example):
    """Test that `ex2rst` does what we think it should
    """
    # render the example to RST with ex2rst
    rst = example.with_suffix(".rst")
    ex2rst.main([str(example), str(rst)])

    # check that it matches
    assert rst.read_text() == EXAMPLE_OUTPUT



# -*- coding: utf-8 -*-
# Copyright (C) Louisiana State University (2014-2017),
#               Cardiff University (2017-2025)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

import inspect
import os.path
import re
import shlex
import shutil
import sys
import warnings
from configparser import ConfigParser
from datetime import date
from pathlib import Path
from string import Template

import matplotlib

from sphinx.util import logging

from numpydoc import docscrape_sphinx

import gwpy
from gwpy.utils.sphinx import (
    ex2rst,
    zenodo,
)

TODAY = date.today()

SPHINX_DIR = Path(__file__).parent.absolute()
STATIC_DIRNAME = "_static"

# use caching in GWpy calls
os.environ.update({
    "GWPY_CACHE": "true",
})

# -- versions ---------------

GWPY_VERSION = gwpy.__version__

# parse version number to get git reference
_setuptools_scm_version_regex = re.compile(
    r"\+g(\w+)(?:\Z|\.)",
)
if match := _setuptools_scm_version_regex.search(GWPY_VERSION):
    GWPY_GIT_REF, = match.groups()
else:
    GWPY_GIT_REF = 'v{}'.format(GWPY_VERSION)

# -- matplotlib -------------

matplotlib.use('agg')

# ignore warnings that aren't useful for documentation
if "CI" not in os.environ:
    for message, category in (
        (".*non-GUI backend.*", UserWarning),
        (".*gwpy.plot.*", DeprecationWarning),
        ("", matplotlib.MatplotlibDeprecationWarning),
    ):
        warnings.filterwarnings("ignore", message=message, category=category)

# -- general ----------------

needs_sphinx = "4.0"
project = "GWpy"
copyright = f"{TODAY.year} Cardiff University"
version = "dev" if ".dev" in GWPY_VERSION else GWPY_VERSION
release = GWPY_VERSION

# extension modules
# DEVNOTE: please make sure and add 3rd-party dependencies to
#          pyproject.toml's [project.optional-dependencies]/docs
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.imgmath',
    'sphinx.ext.autosummary',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.linkcode',
    'sphinx.ext.ifconfig',
    'sphinx_automodapi.automodapi',
    'sphinxcontrib.programoutput',
    'numpydoc',
    'matplotlib.sphinxext.plot_directive',
    #'sphinxcontrib.doxylink',  # noqa: E265
]

# content management
default_role = "obj"
exclude_patterns = [
    "references.rst",
    "_build",
    "_generated",
    "cli/examples/examples.rst",
]
templates_path = [
    "_templates",
]

# epilog
rst_epilog = "\n.. include:: /references.rst"

# -- HTML formatting --------

extensions.append("sphinx_immaterial")
html_theme = "sphinx_immaterial"
html_theme_options = {
    # metadata
    "repo_name": "GWpy",
    "repo_type": "gitlab",
    "repo_url": "https://gitlab.com/gwpy/gwpy",
    "edit_uri": "blob/main/docs",
    "globaltoc_collapse": True,
    # features
    "features": [
        "navigation.sections",
    ],
    # colouring and light/dark mode
    "palette": [
        {
            "media": "(prefers-color-scheme: light)",
            "scheme": "default",
            "primary": "blue-grey",
            "accent": "deep-orange",
            "toggle": {
                "icon": "material/eye-outline",
                "name": "Switch to dark mode",
            },
        },
        {
            "media": "(prefers-color-scheme: dark)",
            "scheme": "slate",
            "primary": "amber",
            "accent": "deep-orange",
            "toggle": {
                "icon": "material/eye",
                "name": "Switch to light mode",
            },
        },
    ],
    # table of contents
    "toc_title_is_page_title": True,
    # version dropdown
    "version_dropdown": True,
    "version_json": "../versions.json",
}
html_static_path = [STATIC_DIRNAME]
html_favicon = str(Path(STATIC_DIRNAME) / "favicon.png")
html_logo = str(Path(STATIC_DIRNAME) / "favicon.png")
html_css_files = ["css/gwpy-sphinx.css"]

# -- extensions config ------

# -- autodoc

autoclass_content = 'class'
autodoc_default_flags = ['show-inheritance', 'members', 'inherited-members']

# -- autosummary

autosummary_generate = True

# -- plot_directive

plot_rcparams = dict(matplotlib.rcParams)
plot_rcparams.update({
    'backend': 'agg',
})
plot_apply_rcparams = True
plot_formats = ['png']
plot_include_source = True
plot_html_show_source_link = False

# -- numpydoc

# fix numpydoc autosummary
numpydoc_show_class_members = False

# use blockquotes (numpydoc>=0.8 only)
numpydoc_use_blockquotes = True

# auto-insert plot directive in examples
numpydoc_use_plots = True

# update the plot detection to include .show() calls
parts = re.split(r'[\(\)|]', docscrape_sphinx.IMPORT_MATPLOTLIB_RE)[1:-1]
parts.extend(('fig.show()', 'plot.show()'))
docscrape_sphinx.IMPORT_MATPLOTLIB_RE = r'\b({})\b'.format('|'.join(parts))

# -- inhertiance_diagram

# configure inheritance diagram
inheritance_graph_attrs = dict(rankdir='TB')

# -- doxylink

LALSUITE_DOCS = 'https://lscsoft.docs.ligo.org/lalsuite'

doxylink = {
    'lal': ('lal.tag', '%s/lal/' % LALSUITE_DOCS),
    'lalframe': ('lalframe.tag', '%s/lalframe/' % LALSUITE_DOCS),
}

# -- intersphinx

# Intersphinx
intersphinx_mapping = {
    'astropy': ('https://docs.astropy.org/en/stable/', None),
    'dateparser': ('https://dateparser.readthedocs.io/en/stable/', None),
    'dateutil': ('https://dateutil.readthedocs.io/en/stable/', None),
    'dqsegdb2': ('https://dqsegdb2.readthedocs.io/en/stable/', None),
    # 'glue': ('https://docs.ligo.org/lscsoft/glue/', None),
    'gwdatafind': ('https://gwdatafind.readthedocs.io/en/stable/', None),
    'gwosc': ('https://gwosc.readthedocs.io/en/stable/', None),
    'h5py': ('https://docs.h5py.org/en/latest/', None),
    'ligo.skymap': ('https://lscsoft.docs.ligo.org/ligo.skymap/', None),
    'igwn-segments': ('https://igwn-segments.readthedocs.io/en/stable/', None),
    'lscsoft-glue': ('https://lscsoft.docs.ligo.org/glue/', None),
    'matplotlib': ('https://matplotlib.org/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'pycbc': ('https://pycbc.org/pycbc/latest/html/', None),
    'python': ('https://docs.python.org/3/', None),
    'uproot': ('https://uproot.readthedocs.io/en/stable/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference/', None),
}


# -- linkcode

def linkcode_resolve(domain, info):
    """Determine the URL corresponding to Python object

    This code is stolen with thanks from the scipy team.
    """
    if domain != "py" or not info["module"]:
        return None

    def find_source(module, fullname):
        obj = sys.modules[module]
        for part in fullname.split("."):
            obj = getattr(obj, part)
        try:  # unwrap a decorator
            obj = obj.im_func.func_closure[0].cell_contents
        except (AttributeError, TypeError):
            pass
        # get filename
        filename = Path(inspect.getsourcefile(obj)).relative_to(
            Path(gwpy.__file__).parent,
        ).as_posix()
        # get line numbers of this object
        source, lineno = inspect.getsourcelines(obj)
        if lineno:
            return "{}#L{:d}-L{:d}".format(
                filename,
                lineno,
                lineno + len(source) - 1,
            )
        return filename

    try:
        fileref = find_source(info["module"], info["fullname"])
    except (
        AttributeError,  # object not found
        OSError,  # file not found
        TypeError,  # source for object not found
        ValueError,  # file not from GWpy
    ):
        return None

    return "https://gitlab.com/gwpy/gwpy/-/tree/{}/gwpy/{}".format(
        GWPY_GIT_REF,
        fileref,
    )


# -- plugins ----------------

# -- build CLI examples

CLI_INDEX_TEMPLATE = Template("""
.. toctree::
   :numbered:

   ${examples}
""".strip())

CLI_TEMPLATE = Template("""
.. _gwpy-cli-example-${tag}:

${titleunderline}
${title}
${titleunderline}

${description}

.. code-block:: shell

   ${command}

.. plot::
   :align: center
   :alt: ${title}
   :context: reset
   :format: python
   :include-source: false

   ${code}
""".strip())


def _new_or_different(content, target):
    """Return `True` if a target file doesn't exist, or doesn't have the
    specified content
    """
    try:
        return Path(target).read_text() != content
    except FileNotFoundError:
        return True


def _render_cli_example(config, section, outdir, logger):
    """Render a :mod:`gwpy.cli` example as RST to be processed by Sphinx.
    """
    # read config values (allow for multi-line definition)
    raw = config.get(
        section,
        'command',
    ).strip().replace("\n", " ") + " --interactive"
    title = config.get(
        section,
        'title',
        fallback=' '.join(map(str.title, section.split('-'))),
    )
    desc = config.get(section, 'description', fallback='')

    # build command-line string for display
    cmdstr = f"gwpy-plot {raw}".replace(  # split onto multiple lines
        ' --',
        ' \\\n       --',
    )

    # build code to generate the plot when sphinx runs
    args = ", ".join(map(repr, shlex.split(raw)))
    code = "\n   ".join([
        "from gwpy.cli.gwpy_plot import main as gwpy_plot",
        f"gwpy_plot([{args}])",
    ])

    rst = CLI_TEMPLATE.substitute(
        title=title,
        titleunderline='#' * len(title),
        description=desc,
        tag=section,
        command=cmdstr,
        code=code,
    )

    # only write RST if new or changed
    rstfile = outdir / f"{section}.rst"
    if _new_or_different(rst, rstfile):
        rstfile.write_text(rst)
        logger.info("[cli] wrote {}".format(rstfile))
    return rstfile


def render_cli_examples(_):
    """Render all :mod:`gwpy.cli` examples as RST to be processed by Sphinx.
    """
    logger = logging.getLogger('cli-examples')

    # directories
    clidir = SPHINX_DIR / "cli"
    exini = clidir / "examples.ini"
    exdir = clidir / "examples"
    exdir.mkdir(exist_ok=True, parents=True)

    # read example config
    config = ConfigParser()
    config.read(exini)

    # render examples
    rsts = []
    for sect in config.sections():
        rst = _render_cli_example(config, sect, exdir, logger)
        # record the path relative to the /cli/ directory
        # because that's where the toctree is included
        rsts.append(rst.relative_to(clidir))

    rst = CLI_INDEX_TEMPLATE.substitute(
        examples="\n   ".join((str(rst.with_suffix("")) for rst in rsts)),
    )
    (exdir / "examples.rst").write_text(rst)


# -- examples

def _render_example(example, outdir, logger):
    # render the example
    rst = ex2rst.ex2rst(example)

    # if it has changed, write it (prevents sphinx from
    # unnecessarily reprocessing)
    target = outdir / example.with_suffix(".rst").name
    if _new_or_different(rst, target):
        target.write_text(rst)
        logger.debug('[examples] wrote {0}'.format(target))


def render_examples(_):
    """Render all examples as RST to be processed by Sphinx.
    """
    logger = logging.getLogger("examples")
    logger.info("[examples] converting examples to RST...")

    srcdir = SPHINX_DIR.parent / "examples"
    outdir = SPHINX_DIR / "examples"
    outdir.mkdir(exist_ok=True)

    # find all examples
    for exdir in next(os.walk(srcdir))[1]:
        if exdir in {"__pycache__"}:  # ignore
            continue
        subdir = outdir / exdir
        subdir.mkdir(exist_ok=True)
        # copy index
        index = subdir / "index.rst"
        shutil.copyfile(srcdir / exdir / index.name, index)
        logger.debug('[examples] copied {0}'.format(index))
        # render python script as RST
        for expy in (srcdir / exdir).glob("*.py"):
            _render_example(expy, subdir, logger)
        logger.info('[examples] converted all in examples/{0}'.format(exdir))


# -- create citation file

def write_citing_rst(app):
    """Render the ``citing.rst`` file using the Zenodo API
    """
    logger = logging.getLogger('zenodo')
    citing = SPHINX_DIR / "citing.rst"
    content = citing.with_suffix(".rst.in").read_text()
    content += '\n' + zenodo.format_citations(597016)
    citing.write_text(content)
    logger.info('[zenodo] wrote {0}'.format(citing))


# -- setup sphinx -----------

def setup(app):
    app.connect('builder-inited', write_citing_rst)
    app.connect('builder-inited', render_examples)
    app.connect('builder-inited', render_cli_examples)



# -*- coding: utf-8 -*-
# Copyright (C) Cardiff University (2023)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Custom filtering utilities for the `TimeSeries`
"""

__author__ = "Alex Southgate <alex.southgate@ligo.org>"

import warnings

import numpy as np
import pytest
import scipy.signal as sig

from ...signal import spectral
from ...timeseries import TimeSeries


@pytest.fixture
def series_data():
    """Create some fake data with equal sampling frequencies.

    Returns
    -------
        firstarr: an array of data, simple mixture of waves
        secondarr: a second array of data from different mixture
        seglen: segment length param to reuse for ffts
    """
    seglen = 512
    n_segs = 10
    n_t = seglen * n_segs
    t_end = 2 * np.pi

    ts = np.linspace(0, t_end, n_t)

    firstarr = 0.1 * np.cos(ts + 0.1) + 0.9 * np.sin(2 * ts + 5)
    firstarr += np.random.normal(5.8, 2, n_t)

    secondarr = 0.5 * np.cos(ts + 0.1) + 0.1 * np.sin(5 * ts + 10)
    secondarr += np.random.normal(5.8, 2, n_t)

    return firstarr, secondarr, seglen


@pytest.fixture
def unequal_fs_series_data():
    """Create some fake data with unequal sampling frequencies.

    Returns
    -------
        ts1: array of time points for first data array
        ts2: array of time points for second data array
        firstarr: an array of data, simple mixture of waves
        secondarr: a second array of data from different mixture
        seglen: segment length param to reuse for ffts
        fs_1: sampling frequency 1
        fs_2: sampling frequency 2
    """
    seglen = 512
    n_segs1 = 10
    n_segs2 = 20
    n_t1 = seglen * n_segs1
    n_t2 = seglen * n_segs2
    t_end = 10

    fs_1 = n_t1 / t_end
    fs_2 = n_t2 / t_end

    ts1 = np.linspace(0, t_end, n_t1)
    ts2 = np.linspace(0, t_end, n_t2)

    firstarr = np.sin(2 * np.pi * ts1) + 0.1 * np.sin(2 * np.pi * ts1 * 5)
    secondarr = np.sin(2 * np.pi * ts2)

    return ts1, ts2, firstarr, secondarr, seglen, fs_1, fs_2


def test_coherence_happy(series_data):
    """Test the interface to scipy.signal.coherence.

    For other tests see timeseries/tests/timeseries.py
    """
    firstarr, secondarr, seglen = series_data
    f_s = 0.001

    first = TimeSeries(firstarr, sample_rate=f_s)
    second = TimeSeries(secondarr, sample_rate=f_s)

    coh = spectral.coherence(first, second, segmentlength=seglen)
    ftemp, cxytemp = sig.coherence(firstarr, secondarr, f_s, nperseg=seglen)

    coharr = np.array(coh.data)

    assert all(coharr == cxytemp)


def test_coherence_resample(unequal_fs_series_data):
    """Ensure warning is raised by unequal sampling frequencies.
    """
    ts1, ts2, firstarr, secondarr, seglen, fs_1, fs_2 = unequal_fs_series_data

    # first and second arrays are different, secondarr should have
    # sampling frequency fs_2, but sometimes a mistake is made
    first = TimeSeries(firstarr, sample_rate=fs_1)
    second = TimeSeries(secondarr, sample_rate=fs_1)
    third = TimeSeries(secondarr, sample_rate=fs_2)

    # the first coherence val coh12 is broken intentionally since
    # secondarr data should not have fs_1, instead fs_2
    coh12 = spectral.coherence(first, second, segmentlength=seglen)
    with pytest.warns(
        UserWarning,
        match="Sampling frequencies are unequal",
    ):
        coh13 = spectral.coherence(first, third, segmentlength=seglen)

    # get the frequency at minimum coherence, this should be the extra
    # component in secondarr
    maxi12 = np.argmin(coh12[:50])
    maxf12 = coh12.frequencies[maxi12]
    maxi13 = np.argmin(coh13[:50])
    maxf13 = coh12.frequencies[maxi13]

    # this one is close to 5 -- the extra freq component in secondarr
    assert 4 <= maxf13.value <= 6
    # this one is totally broken
    assert not (4 <= maxf12.value <= 6)


def test_coherence_resample_downsample(series_data):
    """Ensure warning is raised by unequal sampling frequencies.
    """
    firstarr, secondarr, seglen = series_data
    f_s = 0.001

    first = TimeSeries(firstarr, sample_rate=f_s)
    second = TimeSeries(secondarr, sample_rate=f_s * 2.32)

    with pytest.warns(
        UserWarning,
        match="Sampling frequencies are unequal",
    ):
        coh1 = spectral.coherence(first, second, segmentlength=seglen)

    # check that forcibly disabling downsample results in an error
    with pytest.raises(ValueError):
        spectral.coherence(
            first,
            second,
            segmentlength=seglen,
            downsample=False,
        )

    # but that accepting downsampling gives you the same result as
    # doing nothing (but doesn't emit a warning)
    with warnings.catch_warnings():
        warnings.simplefilter("error", DeprecationWarning)
        coh2 = spectral.coherence(
            first,
            second,
            segmentlength=seglen,
            downsample=True,
        )

    assert all(np.array(coh1.data) == np.array(coh2.data))



# -*- coding: utf-8 -*-
# Copyright (C) Louisiana State University (2014-2017)
#               Cardiff University (2017-2021)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Convert GWpy example python files into rst files for sphinx documentation
"""

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org'

import argparse
import sys
from pathlib import Path

METADATA = {
    # replacement for __metadata__ variables, use None to ignore
    'author': 'sectionauthor',
    'credits': None,
}


def postprocess_code(code, context):
    if any('plot.show()' in line for line in code):
        ctx = "close-figs"
    else:
        code.insert(2, '   :nofigs:')
        ctx = ""

    code.insert(2, '   :context: {}'.format(context).rstrip())
    code.append('')

    return code, ctx


def ex2rst(infile):
    """Convert a Python example script into RST

    Returns
    -------
    rst : `str`
        the fully rendered RST text block
    """
    infile = Path(infile)
    lines = infile.read_text().splitlines()
    ref = '-'.join((infile.parent.name, infile.with_suffix("").name))

    output = []
    header = ['.. _gwpy-example-%s:\n' % ref]

    indoc = False
    incode = False
    code = []
    context = "reset"

    for i, line in enumerate(lines):
        # skip file header
        if len(output) == 0 and line.startswith('#'):
            continue

        # hide lines
        if line.endswith('# hide'):
            continue

        # find block docs
        if '"""' in line:
            indoc = not indoc
            line = line.strip('"')
            if not line:  # start/end of block quote
                continue

        # skip empty lines not in a block quote
        if not line and not indoc:
            if output:
                output.append('')
            continue

        # finish code block
        if incode and line.startswith(('"', '#', '__')):
            incode = False
            code, context = postprocess_code(code, context)
            output.extend(code)

        # comments
        if line.startswith('#'):
            output.append(line[2:])
        # metadata
        elif line.startswith('__'):
            key, value = map(lambda x: x.strip(' _="\'').rstrip(' _="\''),
                             line.split('=', 1))
            try:
                metakey = METADATA[key]
            except KeyError:
                header.append('.. %s:: %s\n' % (key, value))
            else:
                if metakey is not None:
                    header.append('.. %s:: %s\n' % (METADATA[key], value))
        # block quote
        elif indoc:
            output.append(line.strip('"').rstrip('"'))
        # code
        else:
            if not incode:  # restart code block
                code = [
                    '',
                    '.. plot::',
                    '   :include-source:',
                    '',
                ]
            code.append('   %s' % line)
            incode = True

        # end block quote
        if line == '"""' and indoc:
            indoc = False
        elif line.endswith('"""') and indoc:
            output.append('')
            indoc = False

        if len(output) == 1:
            output.append('#'*len(output[0]))

    if incode:
        output.extend(postprocess_code(code, context)[0])

    output = header + output
    return '\n'.join(output).replace('\n\n\n', '\n\n')


def create_parser():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "infile",
        type=Path,
        metavar="example.py",
        help="python file to convert",
    )
    parser.add_argument(
        "outfile",
        type=Path,
        metavar="example.rst",
        nargs="?",
        help="rst file to write, default: print to screen",
    )
    return parser


def main(args=None):
    # parse command line
    parser = create_parser()
    args = parser.parse_args(args)

    # convert python to RST
    rst = ex2rst(args.infile)

    # write output
    if args.outfile:
        with open(args.outfile, "w") as f:
            print(rst, file=f)
    else:
        print(rst)


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014-2020)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Create, manipulate, read, and write time-series data
"""

from .core import (TimeSeriesBase, TimeSeriesBaseDict, TimeSeriesBaseList)
from .timeseries import (TimeSeries, TimeSeriesDict, TimeSeriesList)
from .statevector import (StateVector, StateVectorDict, StateVectorList,
                          StateTimeSeries, StateTimeSeriesDict, Bits)

from . import io

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014-2020)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.time`
"""

from datetime import datetime
from decimal import Decimal
from operator import attrgetter

import pytest

import numpy

from astropy.time import Time
from astropy.units import (UnitConversionError, Quantity)

from ... import time
from .. import LIGOTimeGPS

try:
    from glue.lal import LIGOTimeGPS as GlueGPS
except ImportError:
    GlueGPS = LIGOTimeGPS
    HAS_GLUE = False
else:
    HAS_GLUE = True

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

GW150914 = LIGOTimeGPS(1126259462, 391000000)
GW150914_DT = datetime(2015, 9, 14, 9, 50, 45, 391000)
FREEZE = '2015-09-14 09:50:45.391'
NOW = 1126259462
TODAY = 1126224017
TOMORROW = 1126310417
YESTERDAY = 1126137617


@pytest.mark.freeze_time(FREEZE)
@pytest.mark.parametrize(("in_", "out"), [
    (1126259462, int(GW150914)),
    (1235635623.7500002, LIGOTimeGPS(1235635623, 750000200)),
    (LIGOTimeGPS(1126259462, 391000000), GW150914),
    ('0', 0),
    ('Jan 1 2017', 1167264018),
    ('Sep 14 2015 09:50:45.391', GW150914),
    ('Oct 30 2016 12:34 CST', 1161887657),
    ((2017, 1, 1), 1167264018),
    (datetime(2017, 1, 1), 1167264018),
    (Time(57754, format='mjd'), 1167264018),
    (Time(57754.0001, format='mjd'), LIGOTimeGPS(1167264026, 640000000)),
    (Quantity(1167264018, 's'), 1167264018),
    (Decimal('1126259462.391000000'), GW150914),
    pytest.param(
        GlueGPS(GW150914.gpsSeconds, GW150914.gpsNanoSeconds),
        GW150914,
        marks=pytest.mark.requires("glue"),
    ),
    (numpy.int32(NOW), NOW),  # fails with lal-6.18.0
    ('now', NOW),
    ('today', TODAY),
    ('tomorrow', TOMORROW),
    ('yesterday', YESTERDAY),
])
def test_to_gps(in_, out):
    """Test that :func:`to_gps` works.
    """
    assert time.to_gps(in_) == out


@pytest.mark.parametrize(("in_", "err"), [
    (Quantity(1, 'm'), UnitConversionError),
    ('random string', ValueError),
])
def test_to_gps_error(in_, err):
    """Test that :func:`gwpy.time.to_gps` errors when it should.
    """
    with pytest.raises(err):
        time.to_gps(in_)


@pytest.mark.parametrize('in_, out', [
    (1167264018, datetime(2017, 1, 1)),
    ('1167264018', datetime(2017, 1, 1)),
    (1126259462.391, datetime(2015, 9, 14, 9, 50, 45, 391000)),
    ('1.13e9', datetime(2015, 10, 27, 16, 53, 3)),
    pytest.param(
        GlueGPS(GW150914.gpsSeconds, GW150914.gpsNanoSeconds),
        GW150914_DT,
        marks=pytest.mark.requires("glue"),
    ),
])
def test_from_gps(in_, out):
    """Test that :func:`gwpy.time.from_gps` works.
    """
    assert time.from_gps(in_) == out


@pytest.mark.parametrize(("in_", "err"), [
    ('test', ValueError),
    (1167264017, ValueError),  # gwpy/gwpy#1021
])
def test_from_gps_error(in_, err):
    """Test that :func:`gwpy.time.from_gps` errors when it should.
    """
    with pytest.raises(err):
        time.from_gps(in_)


@pytest.mark.freeze_time(FREEZE)
@pytest.mark.parametrize('in_, out', [
    (float(GW150914), GW150914_DT),
    (GW150914, GW150914_DT),
    (GW150914_DT, GW150914),
    pytest.param(
        GlueGPS(float(GW150914)),
        GW150914_DT,
        marks=pytest.mark.requires("glue"),
    ),
    ('now', NOW),
    ('today', TODAY),
    ('tomorrow', TOMORROW),
    ('yesterday', YESTERDAY),
])
def test_tconvert(in_, out):
    """Test :func:`gwpy.time.tconvert`
    """
    assert time.tconvert(in_) == out


@pytest.mark.parametrize('gpstype', time.gps_types,
                         ids=attrgetter('__module__'))
def test_gps_types(gpstype):
    assert gpstype.__name__ == 'LIGOTimeGPS'
    gps = gpstype(123, 456000000)
    assert gps.gpsSeconds == 123
    assert gps.gpsNanoSeconds == 456000000
    assert gps == 123.456



# -*- coding: utf-8 -*-
# Copyright (C) Cardiff University (2018-2023)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import sys

import requests

DEFAULT_ZENODO_URL = "https://zenodo.org"
DEFAULT_HITS = 10


def format_citations(
    zid,
    url=DEFAULT_ZENODO_URL,
    hits=10,
    tag_prefix="v",
):
    """Query and format a citations page from Zenodo entries.

    Parameters
    ----------
    zid : `int`, `str`
        The Zenodo ID (``conceptrecid``) of the parent target record.

    url : `str`, optional
        The base URL of the Zenodo host, defaults to ``https://zenodo.org``.

    hits : `int`, optional
        The maximum number of results to show, default: ``10``.

    tag_prefix : `str`, optional
        The prefix for git tags. This is removed to generate the section
        headers in the output RST.

    Returns
    -------
    rst : `str`
        An RST-formatted string of DOI badges with URLs.
    """
    # query for metadata
    apiurl = f"{url.rstrip('/')}/api/records"
    params = {
        "q": f"conceptrecid:{zid}",
        "allversions": True,
        "sort": "version",
        "page": 1,
        "size": int(hits),
    }
    resp = requests.get(  # make the request
        apiurl,
        params=params,
        timeout=30,
    )
    resp.raise_for_status()  # make sure it worked
    records = resp.json()  # parse the response

    lines = []
    for i, rec in enumerate(records["hits"]["hits"]):
        # print RST-format header
        version = str(rec['metadata']['version'])[len(tag_prefix):]
        head = "-" * len(version)
        lines.extend([
            head,
            version,
            head,
            "",
        ])

        # add DOI badge
        badge = f"{url}/badge/doi/{rec['doi']}.svg"
        lines.extend([
            f".. image:: {badge}",
            f"    :alt: {rec['title']} Zenodo DOI badge",
            f"    :target: {rec['doi_url']}",
        ])

        # add break before next record
        lines.append("")

    return '\n'.join(lines).strip()


# -- command-line usage ---------------

def create_parser():
    """Create an `argparse.ArgumentParser` for this tool.
    """
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "id",
        type=int,
        help="Zenodo concept ID for package",
    )
    parser.add_argument(
        "-u",
        "--url",
        default=DEFAULT_ZENODO_URL,
        help="Base URL of API to query",
    )
    parser.add_argument(
        "-n",
        "--hits",
        default=DEFAULT_HITS,
        type=int,
        help="Number of versions to display",
    )
    parser.add_argument(
        "-p",
        "--tag-prefix",
        default="v",
        help="Prefix for version tags",
    )
    parser.add_argument(
        "-o",
        "--output-file",
        default="stdout",
        help="Output file path",
    )
    return parser


def main(args=None):
    """Run this tool as a command-line script.
    """
    # parse arguments
    parser = create_parser()
    opts = parser.parse_args(args=args)

    # generate RST
    citing = format_citations(
        opts.id,
        url=opts.url,
        hits=opts.hits,
        tag_prefix=opts.tag_prefix,
    )

    # print
    if opts.output_file in {None, "stdout"}:
        f = sys.stdout
    else:
        f = open(opts.output_file, 'w')
    with f:
        print(citing, file=f)


if __name__ == '__main__':
    sys.exit(main())



"""Run rstcheck and report output for codeclimate."""

from __future__ import annotations

import hashlib
import json
import re
import sys
import typing
from pathlib import Path

from rstcheck_core.config import RstcheckConfig
from rstcheck_core.runner import RstcheckMainRunner

if typing.TYPE_CHECKING:
    from typing import Literal

    from rstcheck_core.types import LintError

MESSAGE_RE = re.compile(
    r"\((?P<severity>[A-Z]+)/[0-9]+\)(\s+)(?P<message>.*)",
)
SEVERITY: dict[str, str] = {
    "ERROR": "major",
    "SEVERE": "major",
    "INFO": "info",
    "WARNING": "minor",
}


class CodeQualityLocation(typing.TypedDict):
    """Location parameter in a `CodeQualityItem`."""

    path: str
    lines: dict[str, int]


class CodeQualityItem(typing.TypedDict):
    """Gitlab Code Quality item.

    See <https://docs.gitlab.com/ci/testing/code_quality/#code-quality-report-format>.
    """

    description: str
    check_name: str
    fingerprint: str
    severity: Literal["info", "minor", "major", "critical", "blocker"]
    location: CodeQualityLocation


def rstcheck(
    config_path: Path,
    check_paths: list[str | Path],
    *,
    recursive: bool = True,
) -> RstcheckMainRunner:
    """Create, execute, and return an `rstcheck_core.runner.RstcheckMainRunner`."""
    config = RstcheckConfig(
        config_path=config_path,
        recursive=recursive,
    )
    runner = RstcheckMainRunner(
        check_paths=list(map(Path, check_paths)),
        rstcheck_config=config,
    )
    runner.check()
    return runner


def codeclimate_item(item: LintError) -> CodeQualityItem:
    """Construct a codeclimate item for this rstcheck lint error."""
    fingerprint = hashlib.sha1(
        "".join(map(str, item.values())).encode("utf-8"),
        usedforsecurity=False,
    ).hexdigest()
    match = MESSAGE_RE.match(item["message"]).groupdict()
    return {
        "check_name": match["message"],
        "description": match["message"],
        "fingerprint": fingerprint,
        "location": {
            "path": str(item["source_origin"]),
            "lines": {
                "begin": (lineno := int(item["line_number"])),
                "end": lineno,
            },
        },
        "severity": SEVERITY.get(match["severity"], "info"),
    }


def codeclimate(items: list[LintError]) -> list[CodeQualityItem]:
    """Parse a list of errors for codeclimate."""
    return [codeclimate_item(item) for item in items]


def main() -> None:
    """Run `rstcheck` and print a codeclimate JSON report."""
    results = rstcheck(
        Path("pyproject.toml"),
        ["docs"],
    )
    issues = codeclimate(results.errors)
    json.dump(issues, sys.stdout)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright (C) Cardiff University (2022)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.io.ffldatafind`
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

import os
from contextlib import nullcontext
from pathlib import Path
from unittest import mock

import pytest

from .. import (
    datafind as io_datafind,
    ffldatafind,
)


# -- test utilities ---------

@pytest.mark.parametrize("content", [
    "abcdef",
    "line 1\nline2",
])
def test_read_last_line(tmp_path, content):
    """Check that the `_read_last_line` utility works
    """
    # write content to a file
    path = tmp_path / "tmp.txt"
    path.write_text(content)
    result = content.rsplit("\n", 1)[-1]
    assert ffldatafind._read_last_line(path) == result


def test_read_last_line_oserror(tmp_path):
    """Check that `_read_last_line` raises `OSError` for empty files.
    """
    path = tmp_path / "tmp.txt"
    path.touch(exist_ok=False)
    with pytest.raises(OSError):
        assert ffldatafind._read_last_line(path)


# -- test ffl utils ---------

@pytest.mark.parametrize(("var", "value", "result"), (
    ("FFLPATH", "/path/to/ffl", "/path/to/ffl"),
    ("VIRGODATA", "/virgoData", "/virgoData/ffl"),
))
@mock.patch.dict("os.environ", clear=True)
def test_get_ffl_basedir(var, value, result):
    """Test that `_get_ffl_basedir` does what it is supposed to.
    """
    # note: use str(Path(x)) to convert to Posix->Windows
    os.environ[var] = str(Path(value))
    assert ffldatafind._get_ffl_basedir() == str(Path(result))


@mock.patch.dict("os.environ", clear=True)
def test_get_ffl_basedir_error():
    """Test that `_get_ffl_basedir` errors when the environment isn't right.
    """
    with pytest.raises(KeyError):
        ffldatafind._get_ffl_basedir()


@pytest.mark.parametrize(("path", "result"), (
    ("test.ffl", True),
    (Path("/path/to/test.ffl"), True),
    ("test.txt", False),
))
def test_is_ffl_file(path, result):
    """Test that `_is_ffl_file` does what it is supposed to.
    """
    assert ffldatafind._is_ffl_file(path) is result


# -- test ffl UI ------------

FFLS = {
    "a/test.ffl": [
        "/tmp/X-test-0-1.gwf 0 1 0 0",
        "/tmp/X-test-1-1.gwf 1 1 0 0",
    ],
    "b/test.ffl": [
        "/tmp/X-test-2-1.gwf 2 1 0 0",
    ],
    "test2.ffl": [
        "/tmp/X-test2-0-1.gwf 0 1 0 0",
        "/tmp/X-test2-1-1.gwf 1 1 0 0",
        "/tmp/X-test2-2-1.gwf 2 1 0 0",
    ],
    "test3.ffl": [
        "/tmp/Y-test3-0-1.gwf 0 1 0 0",
        "/tmp/Y-test3-1-1.gwf 1 1 0 0",
        "/tmp/Y-test3-2-1.gwf 2 1 0 0",
    ],
    "test-empty.ffl": [],
    "test-bad.ffl": ["badness"],
}
TEST_URLS = [
    x.split()[0]
    for key in ("a/test.ffl", "b/test.ffl")
    for x in FFLS[key]
]


@pytest.fixture(autouse=True)
def mock_ffl(tmp_path):
    """Create a temporary FFL file structure and mock it into
    the test environment.
    """
    for path, lines in FFLS.items():
        ffl = tmp_path / path
        ffl.parent.mkdir(parents=True, exist_ok=True)
        ffl.write_text("\n".join(lines))
    with mock.patch.dict(
        "os.environ",
        {"FFLPATH": str(tmp_path)},
        clear=True,
    ):
        yield


@pytest.mark.parametrize(("site", "match", "result"), [
    (None, None, ["test", "test2", "test3"]),
    ("X", None, ["test", "test2"]),
    (None, "2", ["test2"]),
    ("Y", "test", ["test3"]),
])
def test_find_types(site, match, result):
    """Check that `ffldatafind.find_types` works.
    """
    assert sorted(ffldatafind.find_types(
        site=site,
        match=match,
    )) == sorted(result)


@pytest.mark.parametrize(("match", "ctx", "result"), (
    (None, nullcontext(), TEST_URLS),
    (r"\-0\-", pytest.warns(UserWarning), TEST_URLS[:1]),
))
def test_find_urls(match, ctx, result):
    """Check that `ffldatafind.find_urls` works.
    """
    with ctx:
        assert sorted(ffldatafind.find_urls(
            "X",
            "test",
            0,
            3,
            match=match,
        )) == result


@pytest.mark.parametrize(("on_gaps", "ctx"), (
    ("ignore", nullcontext()),
    ("warn", pytest.warns(UserWarning)),
    ("raise", pytest.raises(RuntimeError)),
))
def test_find_urls_on_gaps(on_gaps, ctx):
    """Check that the ``on_gaps`` keyword in `ffldatafind.find_urls`
    works in each case.
    """
    with ctx:
        assert ffldatafind.find_urls(
            "X",
            "test",
            100,
            101,
            on_gaps=on_gaps,
        ) == []


def test_find_latest():
    """Check that `ffldatafind.find_latest` works.
    """
    assert ffldatafind.find_latest(
        "X",
        "test",
    ) == sorted(x.split()[0] for x in FFLS["b/test.ffl"])[-1:]


@pytest.mark.parametrize(("on_missing", "ctx"), (
    ("ignore", nullcontext()),
    ("warn", pytest.warns(UserWarning)),
    ("raise", pytest.raises(RuntimeError)),
))
def test_find_latest_on_missing(on_missing, ctx):
    """Check that the ``on_missing`` keyword in `ffldatafind.find_latest`
    works in each case.
    """
    with ctx:
        assert ffldatafind.find_latest(
            "BAD",
            "BAD",
            on_missing=on_missing,
        ) == []


# -- test gwpy.io.datafind interface

@mock.patch(
    "gwpy.io.datafind.iter_channel_names",
    mock.MagicMock(return_value=["Y1:TEST-CHANNEL"]),
)
@mock.patch(
    "gwpy.io.datafind.on_tape",
    mock.MagicMock(return_value=False),
)
@mock.patch(
    'gwpy.io.datafind.num_channels',
    mock.MagicMock(return_value=1),
)
def test_datafind_find_frametype():
    """Test that gwpy.io.datafind.find_frametype ends up calling out
    to ffldatafind under the right circumstances.
    """
    assert io_datafind.find_frametype(
        "Y1:TEST-CHANNEL",
        allow_tape=True,
    ) == "test3"



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2017-2020)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""I/O utilities for reading `TimeSeries` from a `list` of file paths.
"""

from io import BytesIO
from math import inf
from os import PathLike

from ...io.cache import (
    FILE_LIKE,
    file_segment,
    read_cache,
    write_cache,
)
from ...segments import Segment

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def preformat_cache(cache, start=None, end=None, sort=file_segment):
    """Preprocess a `list` of file paths for reading.

    This function does the following:

    - read the list of paths cache file (if necessary),
    - sort the cache in time order (if possible),
    - sieve the cache to only include data we need.

    Parameters
    ----------
    cache : `list`, `str`, `pathlib.Path`
        List of file paths, or path to a cache file.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS start time of required data, defaults to start of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS end time of required data, defaults to end of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    sort : `callable`, optional
        A callable key function by which to sort the file paths.

    Returns
    -------
    modcache : `list`
        A parsed, sieved list of paths based on the input arguments.

    See also
    --------
    gwpy.io.cache.read_cache
        For details of how the sorting and sieving is implemented
    """
    # if given a list of paths, write it to a file-like structure
    # so that we can use read_cache to do all the work
    if not isinstance(cache, (str, PathLike) + FILE_LIKE):
        cachef = BytesIO()
        write_cache(cache, cachef)
        cachef.seek(0)
        cache = cachef

    # need start and end times to sieve the cache
    if start is None:
        start = -inf
    if end is None:
        end = +inf

    # read the cache
    return read_cache(
        cache,
        coltype=type(start),
        sort=sort,
        segment=Segment(start, end),
    )



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014-2020)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit test for utils module
"""

import platform
import shutil
import subprocess
import sys

import pytest

from .. import shell

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


def test_shell_call():
    with pytest.warns(DeprecationWarning):
        out, err = shell.call([sys.executable, "--version"])
    assert out.rstrip() == 'Python {}'.format(platform.python_version())
    assert err.rstrip() == ''


def test_shell_call_shell():
    with pytest.warns(DeprecationWarning):
        out, err = shell.call("echo This works")
    assert out.rstrip() == "This works"
    assert err.rstrip() == ""


def test_shell_call_error_no_exe():
    with pytest.raises(OSError), pytest.warns(DeprecationWarning):
        shell.call(['this-command-doesnt-exist'])


def test_shell_call_error_shell_no_exe():
    with pytest.raises(subprocess.CalledProcessError), \
         pytest.warns(DeprecationWarning):
        shell.call('this-command-doesnt-exist')


def test_shell_call_error_shell_failure():
    with pytest.raises(subprocess.CalledProcessError), \
         pytest.warns(DeprecationWarning):
        shell.call('python --blah')


def test_shell_call_error_warn():
    with pytest.warns(Warning) as rec:
        shell.call('python --blah', on_error='warn')
    assert len(rec) == 2
    assert rec[0].category is DeprecationWarning
    assert rec[1].category is UserWarning


def test_which():
    with pytest.warns(DeprecationWarning):
        assert shell.which('python') == shutil.which('python')
    with pytest.raises(ValueError), pytest.warns(DeprecationWarning):
        shell.which('gwpy-no-executable')



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014-2020)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.io.datafind`
"""

import os
from unittest import mock

import pytest

import gwdatafind

from ...testing.errors import (
    pytest_skip_network_error,
)
from ...testing.utils import (
    TEST_GWF_FILE,
)
from .. import datafind as io_datafind

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

GWDATAFIND_PATH = "gwpy.io.datafind.gwdatafind"
MOCK_ENV = {
    'VIRGODATA': 'tmp',
    'GWDATAFIND_SERVER': 'test:80',
}


def _mock_gwdatafind(func):
    """Decorate a function to use a mocked GWDataFind API.
    """
    @mock.patch.dict("os.environ", MOCK_ENV)
    @mock.patch(
        f"{GWDATAFIND_PATH}.find_types",
        mock.MagicMock(
            return_value=[os.path.basename(TEST_GWF_FILE).split("-")[1]],
        )
    )
    @mock.patch(
        f"{GWDATAFIND_PATH}.find_urls",
        mock.MagicMock(return_value=[TEST_GWF_FILE]),
    )
    @mock.patch(
        f"{GWDATAFIND_PATH}.find_latest",
        mock.MagicMock(return_value=[TEST_GWF_FILE]),
    )
    @mock.patch(
        'gwpy.io.datafind.iter_channel_names',
        mock.MagicMock(return_value=['L1:LDAS-STRAIN', 'H1:LDAS-STRAIN']),
    )
    @mock.patch(
        'gwpy.io.datafind.num_channels',
        mock.MagicMock(return_value=1),
    )
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


# -- tests --------------------------------------------------------------------

@mock.patch.dict("os.environ", clear=True)
def test_gwdatafind_module_error():
    """Test that _gwdatafind_module() will propagate an exception if
    it can't work out how to find data.
    """
    with pytest.raises(
        RuntimeError,
        match="^unknown datafind configuration",
    ):
        io_datafind.find_frametype("TEST")


# -- find_frametype

@_mock_gwdatafind
def test_find_frametype():
    """Test that `find_frametype` works with basic input.
    """
    assert io_datafind.find_frametype(
        'L1:LDAS-STRAIN',
        allow_tape=True,
    ) == 'HW100916'


@_mock_gwdatafind
def test_find_frametype_return_all():
    """Test that the ``return_all` keyword for `find_frametype` returns a list.
    """
    assert io_datafind.find_frametype(
        'L1:LDAS-STRAIN',
        return_all=True,
    ) == ['HW100916']


@_mock_gwdatafind
def test_find_frametype_multiple():
    """Test that `find_frametype` can handle channels as MIMO.
    """
    assert io_datafind.find_frametype(
        ['H1:LDAS-STRAIN', 'L1:LDAS-STRAIN'],
        allow_tape=True,
    ) == {
        'H1:LDAS-STRAIN': 'HW100916',
        'L1:LDAS-STRAIN': 'HW100916',
    }


@_mock_gwdatafind
def test_find_frametype_error_not_found():
    """Test that `find_frametype` raises the right error for a missing channel.
    """
    # test missing channel raises sensible error
    with pytest.raises(
        ValueError,
        match=(
            r"^Cannot locate the following channel\(s\) "
            "in any known frametype:\n    X1:TEST$"
        )
    ):
        io_datafind.find_frametype('X1:TEST', allow_tape=True)


@_mock_gwdatafind
def test_find_frametype_error_bad_channel():
    """Test that `find_frametype` raises the right error when the channel
    name isn't parsable.
    """
    with pytest.raises(
        ValueError,
        match=(
            "^Cannot parse interferometer prefix from channel name "
            r"'bad channel name', cannot proceed with find\(\)$"
        ),
    ):
        io_datafind.find_frametype('bad channel name')


@_mock_gwdatafind
def test_find_frametype_error_files_on_tape():
    """Test that `find_frametype` raises the right error when the only
    discovered data are on tape, and we asked for not on tape.
    """
    # check that allow_tape errors get handled properly
    patch = mock.patch('gwpy.io.datafind.on_tape', return_value=True)
    raises = pytest.raises(
        ValueError,
        match=r"\[files on tape have not been checked",
    )
    with patch, raises:
        io_datafind.find_frametype('X1:TEST', allow_tape=False)


# -- find_best_frametype

@_mock_gwdatafind
def test_find_best_frametype():
    """Test that `find_best_frametype` works in general.
    """
    assert io_datafind.find_best_frametype(
        'L1:LDAS-STRAIN',
        968654552,
        968654553,
    ) == 'HW100916'


@pytest.mark.requires("LDAStools.frameCPP")
@pytest_skip_network_error
@pytest.mark.skipif(
    "GWDATAFIND_SERVER" not in os.environ,
    reason='No GWDataFind server configured on this host',
)
@pytest.mark.parametrize('channel, expected', [
    ('H1:ISI-GND_STS_ITMY_X_BLRMS_30M_100M.mean,s-trend', 'H1_T'),
    ('H1:ISI-GND_STS_ITMY_X_BLRMS_30M_100M.mean,m-trend', 'H1_M'),
])
def test_find_best_frametype_ligo_trend(channel, expected):
    """Test that `find_best_frametype` correctly matches trends.

    Currently the LIGO trend data are only available from a restricted server,
    so this test only works on LIGO computing resources.
    """
    # check that this server knows about trends and we can authenticate
    try:
        assert expected in gwdatafind.find_types("H"), (
            f"gwdatafind server doesn't know about {expected}"
        )
    except (
        AssertionError,
        RuntimeError,
    ) as exc:
        pytest.skip(str(exc))

    assert io_datafind.find_best_frametype(
        channel,
        1262276680,  # GW200105 -4s
        1262276688,  # GW200105 +4s
    ) == expected


# -- find_latest

@_mock_gwdatafind
def test_find_latest():
    assert io_datafind.find_latest(
        "HLV",
        "HW100916",
    ) == TEST_GWF_FILE


@mock.patch(
    f"{GWDATAFIND_PATH}.find_latest",
    mock.MagicMock(side_effect=IndexError),
)
@_mock_gwdatafind
def test_find_latest_error():
    with pytest.raises(RuntimeError, match="^no files found for X-MISSING$"):
        io_datafind.find_latest("X1", "MISSING")


# -- find_types

@mock.patch.dict("os.environ", MOCK_ENV)
def test_find_types():
    """Check that `find_types` works with default arguments.
    """
    types = ["a", "b", "c"]
    with mock.patch(f"{GWDATAFIND_PATH}.find_types", return_value=types):
        assert io_datafind.find_types(
            "X",
        ) == sorted(types)


@mock.patch.dict("os.environ", MOCK_ENV)
def test_find_types_priority():
    """Check that `find_types` prioritises trends properly.
    """
    types = ["L1_R", "L1_T", "L1_M"]
    with mock.patch(f"{GWDATAFIND_PATH}.find_types", return_value=types):
        assert io_datafind.find_types(
            "X",
            trend="m-trend",
        ) == ["L1_M", "L1_R", "L1_T"]


# -- utilities --------------

def test_on_tape_false():
    """Check `on_tape` works with a normal file.
    """
    assert io_datafind.on_tape(TEST_GWF_FILE) is False


def test_on_tape_true():
    """Test that `on_tape` returns `True` for a file with zero blocks.
    """
    with mock.patch("os.stat") as os_stat:
        os_stat.return_value.st_blocks = 0
        assert io_datafind.on_tape(TEST_GWF_FILE) is True


def test_on_tape_windows():
    """Test that `on_tape` always returns `False` on Windows.
    """
    class _Stat():
        def __init__(self, path):
            pass

        @property
        def st_blocks(self):
            raise AttributeError("st_blocks")

    with mock.patch("os.stat", _Stat):
        assert io_datafind.on_tape(TEST_GWF_FILE) is False


@pytest.mark.parametrize('ifo, ftype, trend, priority', [
    ('L1', 'L1_HOFT_C00', None, 1),  # hoft
    ('H1', 'H1_HOFT_C02_T1700406_v3', None, 1),  # cleaned hoft
    ('H1', 'H1_M', 'm-trend', 0),  # minute trends
    ('K1', 'K1_T', 's-trend', 0),  # second trends
    ('K1', 'K1_R', 's-trend', 5),  # raw type when looking for second trend
    ('K1', 'K1_M', None, 10),  # trend type, but not looking for trend channel
    ('K1', 'K1_C', None, 6),  # commissioning type
    ('X1', 'SOMETHING_GRB051103', None, 10),  # low priority type
    ('X1', 'something else', None, 5),  # other
])
def test_type_priority(ifo, ftype, trend, priority):
    """Test that `_type_priority` works for various cases.
    """
    assert io_datafind._type_priority(ifo, ftype, trend=trend)[0] == priority



# -*- coding: utf-8 -*-
# Copyright (C) Louisiana State University (2014-2017)
#               Cardiff University (2017-2021)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Support for plotting with units
"""

from astropy.units.format import LatexInline


class LatexInlineDimensional(LatexInline):
    r"""Custom LaTeX formatter that includes physical type (if available)

    Mainly for auto-labelling `Axes` in matplotlib figures.

    Examples
    --------
    The built-in astropy ``latex_inline`` formatter gives this:

    >>> Unit('m/s').to_string(format='latex_inline')
    '$\mathrm{m\,s^{-1}}$'

    This custom 'dimensional' formatter gives:

    >>> Unit('m/s').to_string(format='latex_inline_dimensional')
    'Speed [$\mathrm{m\,s^{-1}}$]'
    """
    name = 'latex_inline_dimensional'

    @classmethod
    def to_string(cls, unit, *args, **kwargs):
        u = f"[{super().to_string(unit, *args, **kwargs)}]"

        if unit.physical_type not in {None, 'unknown', 'dimensionless'}:
            # format physical type of unit for LaTeX
            ptype = str(unit.physical_type).title().replace("_", r"\_")
            # return '<Physical type> [<unit>]'
            return f"{ptype} {u}"
        return u



# -*- coding: utf-8 -*-
# Copyright (C) California Institute of Technology (2022)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.table.io.gstlal`
"""

import pytest

from numpy import testing as nptest
from numpy import float32

from ..io import gstlal as gstlalio
from gwpy.table import EventTable

__author__ = 'Derek Davis <derek.davis@ligo.org>'


# -- gstlal file fixture -----------------------------------------------------


GSTLAL_FILE = """<?xml version='1.0' encoding='utf-8'?>
<!DOCTYPE LIGO_LW SYSTEM "http://ldas-sw.ligo.caltech.edu/doc/ligolwAPI/html/ligolw_dtd.txt">
<LIGO_LW>
        <Table Name="coinc_inspiral:table">
                <Column Name="coinc_event:coinc_event_id" Type="int_8s"/>
                <Column Name="combined_far" Type="real_8"/>
                <Column Name="end_time" Type="int_4s"/>
                <Column Name="end_time_ns" Type="int_4s"/>
                <Column Name="false_alarm_rate" Type="real_8"/>
                <Column Name="ifos" Type="lstring"/>
                <Column Name="mass" Type="real_8"/>
                <Column Name="mchirp" Type="real_8"/>
                <Column Name="minimum_duration" Type="real_8"/>
                <Column Name="snr" Type="real_8"/>
                <Stream Name="coinc_inspiral:table" Delimiter="," Type="Local">
                        1,1,100,0,1,"H1,L1",1,1,1,1,
                        2,1,100,0,1,"H1,L1",1,1,1,1,
                </Stream>
        </Table>
        <Table Name="coinc_event:table">
                <Column Name="coinc_definer:coinc_def_id" Type="int_8s"/>
                <Column Name="coinc_event_id" Type="int_8s"/>
                <Column Name="instruments" Type="lstring"/>
                <Column Name="likelihood" Type="real_8"/>
                <Column Name="nevents" Type="int_4u"/>
                <Column Name="process:process_id" Type="int_8s"/>
                <Column Name="time_slide:time_slide_id" Type="int_8s"/>
                <Stream Name="coinc_event:table" Delimiter="," Type="Local">
                        0,1,"H1,L1",1,1,0,0,
                        1,1,"H1,L1",1,1,0,0
                </Stream>
        </Table>
        <Table Name="sngl_inspiral:table">
                <Column Name="process:process_id" Type="int_8s"/>
                <Column Name="ifo" Type="lstring"/>
                <Column Name="end_time" Type="int_4s"/>
                <Column Name="end_time_ns" Type="int_4s"/>
                <Column Name="eff_distance" Type="real_4"/>
                <Column Name="coa_phase" Type="real_4"/>
                <Column Name="mass1" Type="real_4"/>
                <Column Name="mass2" Type="real_4"/>
                <Column Name="snr" Type="real_4"/>
                <Column Name="chisq" Type="real_4"/>
                <Column Name="chisq_dof" Type="int_4s"/>
                <Column Name="bank_chisq" Type="real_4"/>
                <Column Name="bank_chisq_dof" Type="int_4s"/>
                <Column Name="sigmasq" Type="real_8"/>
                <Column Name="spin1x" Type="real_4"/>
                <Column Name="spin1y" Type="real_4"/>
                <Column Name="spin1z" Type="real_4"/>
                <Column Name="spin2x" Type="real_4"/>
                <Column Name="spin2y" Type="real_4"/>
                <Column Name="spin2z" Type="real_4"/>
                <Column Name="template_duration" Type="real_8"/>
                <Column Name="event_id" Type="int_8s"/>
                <Column Name="Gamma0" Type="real_4"/>
                <Column Name="Gamma1" Type="real_4"/>
                <Column Name="Gamma2" Type="real_4"/>
                <Stream Name="sngl_inspiral:table" Delimiter="," Type="Local">
                        1,"L1",100,0,nan,0,10,10,3,2,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,
                        1,"H1",100,0,nan,0,10,10,3,2,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,
                        1,"V1",100,0,nan,0,10,10,3,2,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1
                </Stream>
        </Table>
</LIGO_LW>
"""  # noqa: E501


@pytest.fixture
def gstlal_table(tmp_path):
    tmp = tmp_path / "H1L1V1-LLOID-1-1.xml.gz"
    tmp.write_text(GSTLAL_FILE)
    return tmp


# -- test data ----------------------------------------------------------------

SNGL_LEN = 3
COINC_LEN = 2


@pytest.mark.requires("ligo.lw.lsctables")
def test_sngl_function(gstlal_table):
    table = gstlalio.read_gstlal_sngl(gstlal_table)
    assert len(table) == SNGL_LEN


@pytest.mark.requires("ligo.lw.lsctables")
def test_read_sngl_eventtable(gstlal_table):
    table = EventTable.read(gstlal_table, format='ligolw.gstlal',
                            triggers='sngl')
    assert len(table) == SNGL_LEN


@pytest.mark.requires("ligo.lw.lsctables")
def test_read_sngl_format(gstlal_table):
    table = EventTable.read(gstlal_table, format='ligolw.gstlal.sngl')
    assert len(table) == SNGL_LEN


@pytest.mark.requires("ligo.lw.lsctables")
def test_read_sngl_columns(gstlal_table):
    table = EventTable.read(gstlal_table, format='ligolw.gstlal.sngl',
                            columns=['snr', 'end_time'])
    assert list(table.keys()) == ['snr', 'end_time']


@pytest.mark.requires("ligo.lw.lsctables")
def test_coinc_function(gstlal_table):
    table = gstlalio.read_gstlal_coinc(gstlal_table)
    assert len(table) == COINC_LEN


@pytest.mark.requires("ligo.lw.lsctables")
def test_read_coinc_eventtable(gstlal_table):
    table = EventTable.read(gstlal_table, format='ligolw.gstlal',
                            triggers='coinc')
    assert len(table) == COINC_LEN


@pytest.mark.requires("ligo.lw.lsctables")
def test_read_coinc_format(gstlal_table):
    table = EventTable.read(gstlal_table, format='ligolw.gstlal.coinc')
    assert len(table) == COINC_LEN


@pytest.mark.requires("ligo.lw.lsctables")
def test_read_coinc_columns(gstlal_table):
    table = EventTable.read(gstlal_table, format='ligolw.gstlal.coinc',
                            columns=['snr', 'end_time'])
    assert list(table.keys()) == ['snr', 'end_time']


@pytest.mark.requires("ligo.lw.lsctables")
def test_derived_values(gstlal_table):
    table = EventTable.read(gstlal_table, format='ligolw.gstlal',
                            triggers='sngl',
                            columns=['snr_chi', 'chi_snr', 'mchirp'])
    nptest.assert_almost_equal(
        table['snr_chi'][0], 4.5)
    nptest.assert_almost_equal(
        table['chi_snr'][0], 1./4.5)
    nptest.assert_almost_equal(
        table['mchirp'][0], float32(8.705506))


@pytest.mark.requires("ligo.lw.lsctables")
def test_incorrect_sngl_column(gstlal_table):
    with pytest.raises(
         ValueError,
         match="is not a valid column name",
         ):
        EventTable.read(gstlal_table, format='ligolw.gstlal.sngl',
                        columns=['nan'])


@pytest.mark.requires("ligo.lw.lsctables")
def test_incorrect_coinc_column(gstlal_table):
    with pytest.raises(
         ValueError,
         match="is not a valid column name",
         ):
        EventTable.read(gstlal_table, format='ligolw.gstlal.coinc',
                        columns=['nan'])


@pytest.mark.requires("ligo.lw.lsctables")
def test_incorrect_trigger_name(gstlal_table):
    with pytest.raises(
         ValueError,
         match="^The 'triggers' argument",
         ):
        EventTable.read(gstlal_table, format='ligolw.gstlal',
                        triggers='nan')



# -*- coding: utf-8 -*-
# Copyright (C) Patrick Godwin (2019-2020)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read events from SNAX
"""

import warnings

from astropy.io.misc.hdf5 import read_table_hdf5
from astropy.table import vstack

from ...io.hdf5 import with_read_hdf5
from ...io.registry import register_reader
from .. import EventTable
from .utils import (read_with_columns, read_with_selection)

__author__ = 'Patrick Godwin <patrick.godwin@ligo.org>'


@read_with_columns
@read_with_selection
@with_read_hdf5
def table_from_file(source, channels=None, on_missing="error", compact=False):
    """Read an `EventTable` from a SNAX HDF5 file

    Parameters
    ----------
    source : `h5py.File`
        the file path of open `h5py` object from which to read the data

    channels : `str` or `list`, optional
        the channel(s) to read data from. if no channels selected,
        read in all channels

    on_missing: `str`, optional
        how to proceed when channels requested are missing. choose one from:
            * ``'warn'``: emit a warning when missing channels are discovered
            * ``'error'``: raise an exception
        default is 'error'.

    compact : `bool`, optional, default: False
        whether to store a compact integer representation in the channel
        column rather than the full channel name, instead storing a mapping
        (`channel_map`) in the table metadata.

    Returns
    -------
    table : `~gwpy.table.EventTable`
    """
    # format channels appropriately
    if isinstance(channels, str):
        channels = [channels]

    # only query channels contained in file
    # if channels not specified, load all of them
    if channels is None:
        channels = source.keys()
    else:
        channels = set(channels)
        found = set(source.keys())
        missing = channels - found
        # check whether missing channels should
        # be an error or simply a warning
        if missing:
            msg = "requested channels not found in SNAX file: '{}'".format(
                "', '".join(missing),
            )
            if on_missing == "error":
                raise ValueError(msg)
            elif on_missing == "warn":
                warnings.warn(msg)
            else:
                raise ValueError(
                    'on_missing argument must be one of "warn" or "error"')
        channels = channels & found

    # read data, adding in 'channel' column
    # to preserve uniqueness across channels
    tables = []
    for channel in channels:
        table = vstack(
            list(map(read_table_hdf5, source[channel].values())),
            join_type="exact",
            metadata_conflicts="error",
        )
        # determine whether to store a compact
        # representation of the channel, storing
        # the mapping to table metadata instead
        if compact:
            table["channel"] = hash(channel)
            table.meta["channel_map"] = {hash(channel): channel}
        else:
            table["channel"] = channel
        tables.append(table)

    # combine results
    return vstack(tables, join_type="exact", metadata_conflicts="error")


# register for unified I/O
register_reader('hdf5.snax', EventTable, table_from_file)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014-2020)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.io.utils`
"""

import gzip
import tempfile
from pathlib import Path

import pytest

from .. import (
    cache as io_cache,
    utils as io_utils,
)

from .test_cache import cache  # noqa: F401

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


def test_gopen(tmp_path):
    tmp = tmp_path / "test.tmp"
    # test simple use
    with tmp.open('w') as f:
        f.write('blah blah blah')
    with io_utils.gopen(tmp) as f2:
        assert f2.read() == 'blah blah blah'


@pytest.mark.parametrize("suffix", (".txt.gz", ""))
def test_gopen_gzip(tmp_path, suffix):
    tmp = tmp_path / f"test{suffix}"
    text = b'blah blah blah'
    with gzip.open(tmp, 'wb') as fobj:
        fobj.write(text)
    with io_utils.gopen(tmp, mode='rb') as fobj2:
        assert isinstance(fobj2, gzip.GzipFile)
        assert fobj2.read() == text


def test_identify_factory():
    id_func = io_utils.identify_factory('.blah', '.blah2')
    assert id_func(None, None, None) is False
    assert id_func(None, 'test.txt', None) is False
    assert id_func(None, 'test.blah', None) is True
    assert id_func(None, 'test.blah2', None) is True
    assert id_func(None, 'test.blah2x', None) is False


def test_file_list_file(cache):  # noqa: F811
    # test file -> [file.name]
    with tempfile.NamedTemporaryFile() as f:
        assert io_utils.file_list(f) == [f.name]


@pytest.mark.requires("lal")
def test_file_list_cache(cache):  # noqa: F811
    from lal.utils import CacheEntry
    # test CacheEntry -> [CacheEntry.path]
    lcache = list(map(CacheEntry.from_T050017, cache))
    assert io_utils.file_list(lcache[0]) == [cache[0]]

    # test cache object -> pfnlist
    assert io_utils.file_list(lcache) == cache

    # test cache file -> pfnlist()
    with tempfile.NamedTemporaryFile(suffix='.lcf', mode='w') as f:
        io_cache.write_cache(lcache, f)
        f.seek(0)
        assert io_utils.file_list(f.name) == cache


def test_file_list_str():
    # test comma-separated list -> list
    assert io_utils.file_list('A,B,C,D') == ['A', 'B', 'C', 'D']

    # test list -> list
    assert io_utils.file_list(['A', 'B', 'C', 'D']) == ['A', 'B', 'C', 'D']


def test_file_list_error():
    with pytest.raises(ValueError):
        io_utils.file_list(1)


@pytest.mark.parametrize(("input_", "expected"), (
    ("test.txt", "test.txt"),
    ("file:///test/path.txt", "/test/path.txt"),
    (Path("test.txt"), "test.txt"),
))
def test_file_path(input_, expected):
    assert io_utils.file_path(input_) == expected


def test_file_path_file():
    """Check that :func:`gwpy.io.utils.file_path` can handle open files
    """
    with tempfile.NamedTemporaryFile() as f:
        assert io_utils.file_path(f) == f.name


@pytest.mark.parametrize("badthing", (
    1,
    ["test.txt"],
))
def test_file_path_errors(badthing):
    """Check that :func:`gwpy.io.utils.file_path` fails when expected
    """
    with pytest.raises(
        ValueError,
        match="^Cannot parse file name for ",
    ):
        io_utils.file_path(badthing)


@pytest.mark.requires("lal")
def test_file_path_cacheentry():
    from lal.utils import CacheEntry
    path = "/path/to/A-B-0-1.txt"
    assert io_utils.file_path(CacheEntry.from_T050017(path)) == path



# Copyright (c) 2017-2025 Cardiff University
#               2014-2017 Louisiana State University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Custom units and formatting."""

import contextlib
import re

from astropy import (
    __version__ as astropy_version,
    units,
)
from astropy.units import imperial as units_imperial
from astropy.units.format.generic import Generic
from packaging.version import Version

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

#: Is the current version of Astropy 7.1 or later?
ASTROPY_71 = Version(astropy_version) >= Version("7.1.0a0")

# container for new units (so that each one only gets created once)
UNRECOGNIZED_UNITS = {}


# -- parser to handle any unit ------------------------------------------------

class GWpyFormat(Generic):
    """Sub-class of the `~astropy.units.format.Generic` unit parser that is forgiving.

    This format tries to work around 'human' errors in unit naming,
    including plurals and capitalisation, and if nothing else works
    it just defines a new unit matching the given string.

    New units are not registered, so cannot be referred to later, but are
    created so that mathematical operations will work. Conversions to other
    units will explicitly not work.
    """

    name = "gwpy"
    _re_closest_unit = re.compile(r"Did you mean (.*)\?\Z")
    _re_closest_unit_delim = re.compile("(, | or )")

    @classmethod
    def _get_unit(cls, t):
        """Get the unit for a lexer token, with lenient matching.

        This method handles the lenient unit parsing for all Astropy versions.
        In Astropy < 7.1, there was no base implementation of this method.
        In Astropy 7.1, this method was removed in favor of `_validate_unit`.
        In Astropy >= 7.2, this method was restored to the base class.
        """
        exc = None
        name = t.value

        # For Astropy >= 7.1, use the base class implementation first
        if ASTROPY_71:
            try:
                return super()._get_unit(t)
            except ValueError as err:
                exc = err
        # For Astropy < 7.1, call _parse_unit directly
        else:
            try:
                return cls._parse_unit(name)
            except ValueError as err:
                exc = err

        # Common fallback logic for all versions
        singular = name[:-1] if name.endswith("s") else ""

        # Parse alternative units from the error message
        # split 'A, B, or C' -> ['A', 'B', 'C']
        if exc and (match := cls._re_closest_unit.search(str(exc))):
            alts = set(cls._re_closest_unit_delim.split(match.groups()[0])[::2])
        else:
            alts = set()

        candidates = list(filter(None, (
            # match uppercase to titled (e.g. MPC -> Mpc)
            name.title(),
            # match titled unit to lower-case (e.g. Amp -> amp)
            name.lower(),
            # match plural to singular (e.g. meters -> meter)
            singular,
            singular.lower() if singular else None,
        )))

        for candidate in candidates:
            if candidate in alts:
                if ASTROPY_71:
                    # Use _validate_unit for Astropy >= 7.1
                    try:
                        return cls._validate_unit(candidate)
                    except (ValueError, KeyError):
                        continue
                else:
                    # Use _parse_unit for Astropy < 7.1
                    return cls._parse_unit(candidate)

        # Re-raise the original exception if no alternatives worked
        if exc:
            raise exc
        msg = f"Unit '{name}' not recognized"
        raise ValueError(msg)


def parse_unit(
    name,
    parse_strict="warn",
    format="gwpy",
):
    """Attempt to intelligently parse a `str` as a `~astropy.units.Unit`.

    Parameters
    ----------
    name : `str`
        unit name to parse

    parse_strict : `str`
        one of 'silent', 'warn', or 'raise' depending on how pedantic
        you want the parser to be

    format : `~astropy.units.format.Base`
        the formatter class to use when parsing the unit string

    Returns
    -------
    unit : `~astropy.units.UnitBase`
        the unit parsed by `~astropy.units.Unit`

    Raises
    ------
    ValueError
        if the unit cannot be parsed and `parse_strict='raise'`
    """
    if name is None or isinstance(name, units.UnitBase):
        return name

    # have we already handled this new unit?
    with contextlib.suppress(KeyError):
        return UNRECOGNIZED_UNITS[name]

    # no, either a valid unit, or something new
    try:
        return units.Unit(name, parse_strict="raise")
    except ValueError as exc:
        if (
            # the format was selected by the user
            format in {None, "generic"}
            # or we were asked to be strict about things
            or parse_strict == "raise"
            # or this isn't the error we're looking for
            or "did not parse as unit" not in str(exc)
        ):
            raise
        # try again using our own lenient parser
        new = units.Unit(name, parse_strict=parse_strict, format=format)
        UNRECOGNIZED_UNITS[name] = new
        return new


# -- custom units -------------------------------------------------------------
# pylint: disable=no-member,invalid-name

# enable imperial units
units.add_enabled_units(units_imperial)

# -- custom units settings
# the following happens in two sets
#     1) alternative names for standard units where SI prefices will not
#        be used
#     2) new units or alternative names for standard units where SI prefices
#        _will_ be used
#
# for developers: when adding a new custom unit, please remember to add it
# to the list of tested units in `test_detector.py`

# 1) alternative names
registry = units.get_current_unit_registry().registry
for unit, aliases in [
    (units.Unit('ct'), ('counts',)),
    (units.Unit('Celsius'), ('Degrees_C', 'DegC')),
    (units.Unit('Fahrenheit'), ('Degrees_F', 'DegF')),
    # GW observatories like to record 'time' as the unit
    (units.Unit('second'), (
        'time',
        'time (s)',
        'time [s]',
        'Time [sec]',
        'Time (sec)',
        'Seconds',  # GWOSC
    )),
]:
    unit.names.extend(aliases)
    for alias in aliases:
        registry[alias] = unit

# 2) new units
_ns = {}

# LIGO-Lab standard for 'no unit defined'
units.def_unit(['NONE', 'undef'], namespace=_ns,
               doc='No unit has been defined for these data')

# other dimenionless units
units.def_unit('strain', namespace=_ns)
units.def_unit('coherence', represents=units.dimensionless_unscaled, namespace=_ns)

# alias for 'second' but with prefices
units.def_unit((['sec'], ['sec']), represents=units.second, prefixes=True,
               namespace=_ns)

# alternative Pressure unit for LIGO UHV
units.def_unit((['torr'], ['torr']), represents=101325/760.*units.pascal,
               prefixes=True, namespace=_ns)

# pounds per square inch gauge
units.def_unit('psig', represents=units_imperial.psi, prefixes=True,
               namespace=_ns, doc='Pound per square inch gauge: pressure')

# cubic feet
units.def_unit('cf', represents=units_imperial.foot**3, namespace=_ns)

# cubic feet per minute
units.def_unit('cfm', represents=_ns['cf']/units.minute, namespace=_ns)

# particles (as in dust)
units.def_unit(['ptcls', 'particles', 'particulates'], prefixes=True,
               namespace=_ns)

# -- register units -----------------------------
units.add_enabled_units(_ns)



# -*- coding: utf-8 -*-
# Copyright (C) Cardiff University (2022)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""GWDataFind UI for FFL cache files.

This module is used to replace the proper GWDataFind interface
on-the-fly when FFL data access is inferred.
As such this module is required to emulate those functions
from `gwdatafind` used in :mod:`gwpy.io.datafind`.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

import os
import re
from collections import defaultdict
from functools import lru_cache
from warnings import warn

from igwn_segments import (
    segment,
    segmentlist,
)

from .cache import (
    _CacheEntry,
    _iter_cache,
    cache_segments,
    file_segment,
    read_cache_entry,
)

_SITE_REGEX = re.compile(r"\A(\w+)-")
_DEFAULT_TYPE_MATCH = re.compile(r"^(?!lastfile|spectro|\.).*")


# -- generic utilities ------

def _read_last_line(path, bufsize=2, encoding="utf-8"):
    """Read the last line of a file.
    """
    with open(path, "rb") as fobj:
        # go to end of file
        fobj.seek(-bufsize, os.SEEK_END)

        # rewind until we hit a line break
        while fobj.read(1) != b"\n":
            try:
                fobj.seek(-bufsize, os.SEEK_CUR)
            except OSError:
                # if we've rewound to the start of the file, just stop
                if fobj.tell() < bufsize:
                    fobj.seek(0)
                    break
                # otherwise this is a different error
                raise

        # read the current line
        return fobj.readline().rstrip().decode(encoding)


# -- ffl utilities ----------

def _get_ffl_basedir():
    """Return the base directory in which to find FFL files

    Raises
    ------
    KeyError
        If neither the ``FFLPATH`` or ``VIRGODATA`` environment variables
        are set.
    """
    if 'FFLPATH' in os.environ:
        return os.environ['FFLPATH']
    if 'VIRGODATA' in os.environ:
        return os.path.join(os.environ['VIRGODATA'], 'ffl')
    raise KeyError(
        "failed to parse FFLPATH from environment, please set "
        "FFLPATH to point to the directory containing FFL files",
    )


def _is_ffl_file(path):
    """Return `True` if this file looks (naively) like an FFL file.
    """
    return str(path).endswith(".ffl")


def _get_site_tag(path):
    """Return the ``(site, tag)`` for a given FFL file.
    """
    # tag is just name of file minus extension
    tag = os.path.splitext(os.path.basename(path))[0]

    # need to read first file from FFL to get site (IFO)
    last = _read_last_line(path).split()[0]
    site = _SITE_REGEX.match(os.path.basename(last)).groups()[0]

    return site, tag


def _find_ffl_files(basedir=None):
    """Find all FFL files under a given base directory.
    """
    for root, _, files in os.walk(basedir or _get_ffl_basedir()):
        for name in filter(_is_ffl_file, files):
            yield os.path.join(root, name)


@lru_cache()
def _find_ffls(basedir=None):
    """Find all readable FFL files.
    """
    ffls = defaultdict(list)
    for path in _find_ffl_files(basedir=basedir):
        try:
            ffls[_get_site_tag(path)].append(path)
        except (
            OSError,  # file is empty (or cannot be read at all)
            AttributeError,  # last entry didn't match _SITE_REGEX
        ):
            continue
    return ffls


def _ffl_paths(site, tag, basedir=None):
    """Return the paths of all FFL files for a given site and tag.
    """
    try:
        return _find_ffls(basedir=basedir)[(site, tag)]
    except KeyError:
        raise ValueError(
            f"no FFL file found for ('{site}', '{tag}')",
        )


@lru_cache()
def _read_ffls(site, tag, basedir=None):
    """Read all FFL files for a given site and tag
    as a list of `CacheEntry` objects.
    """
    entries = []
    for ffl in _ffl_paths(site, tag, basedir=basedir):
        with open(ffl, "r") as fobj:
            entries.extend(
                _CacheEntry(site, tag, entry.segment, entry.path)
                for entry in _iter_cache(fobj, gpstype=float)
            )
    return entries


def _handle_error(action, message):
    """Handle error, warn, or ignore for the given state.

    Parameters
    ----------
    action : `str`
        The action to perform, one of
        ``'warn'`` (emit a `UserWarning`, default),
        ``'ignore'`` (do nothing),
        or anything else (raise a `RuntimeError`).

    message : `str`
        The message to emit with warnings or errors.

    Raises
    ------
    RuntimeError
        If action is not ``'warn'`` or ``'ignore'``.
    """
    # if ignore, do nothing
    if action == "ignore":
        return
    # if warn, emit a warning
    if action == "warn":
        warn(message, stacklevel=2)
        return

    # otherwise, raise an error
    raise RuntimeError(message)


# -- ui ---------------------

def find_types(site=None, match=_DEFAULT_TYPE_MATCH):
    """Return the list of known data types.

    Parameters
    ----------
    site : `str`, optional
        Observatory ID (e.g. ``'A'`)) to retrict types, if `None` (default)
        is given, all types are returned.

    match : `str`, optional
        Regular expression to use to restrict types.

    Returns
    -------
    types : `list` of `str`
        The list of data types matching the criteria.
    """
    ffls = _find_ffls()
    types = [tag for (site_, tag) in ffls if site in (None, site_)]
    if match is not None:
        match = re.compile(match)
        return list(filter(match.search, types))
    return types


def find_urls(
    site,
    tag,
    gpsstart,
    gpsend,
    match=None,
    on_gaps="warn",
):
    """Return the list of all files of the given type in the [start, end)
    GPS interval.

    Parameters
    ----------
    site : `str`
        Observatory ID to search for.

    tag : `str`
        Data type tag to search for.

    gpsstart : `int`
        GPS start time of query.

    gpsend : `int`
        GPS end time of query.

    match : `str`, optional
        Regular expression to use to retrict returned data URLs.

    on_gaps : `str`, optional
        What to do if the full GPS interval is not covered, one of
        ``'warn'`` (emit a `UserWarning`, default),
        ``'ignore'`` (do nothing),
        anything else (raise a `RuntimeError`).

    Returns
    -------
    urls : `list` of `str`
        A list of URLs representing discovered data.
    """
    if match:
        match = re.compile(match)

    span = segment(gpsstart, gpsend)

    cache = [
        e for e in _read_ffls(site, tag) if (
            e.observatory == site
            and e.description == tag
            and e.segment.intersects(span)
            and (match.search(e.path) if match else True)
        )
    ]
    urls = [e.path for e in cache]

    # handle missing data
    missing = segmentlist([span]) - cache_segments(cache)
    if missing:
        _handle_error(
            on_gaps,
            "Missing segments: \n" + "\n".join(map(str, missing)),
        )

    return urls


def find_latest(site, tag, on_missing="warn"):
    """Return the most recent file of a given type.

    Parameters
    ----------
    site : `str`
        Observatory ID to search for.

    tag : `str`
        Data type tag to search for.

    on_missing : `str`, optional
        What to do if a URL is not found for the given site and tag, one of
        ``'warn'`` (emit a `UserWarning`, default),
        ``'ignore'`` (do nothing),
        anything else (raise a `RuntimeError`).

    Returns
    -------
    urls : `list` of `str`
        A list (typically of one item) of URLs representing the latest data
        for a specific site and tag.
    """
    try:
        fflfiles = _ffl_paths(site, tag)
    except ValueError:  # no readable FFL file
        urls = []
    else:
        urls = [
            read_cache_entry(_read_last_line(fflfile), gpstype=float)
            for fflfile in fflfiles
        ]
        if urls:  # if multiple, find the latest one
            urls = sorted(urls, key=file_segment)[-1:]

    if not urls:
        _handle_error(on_missing, "No files found")

    return urls



# -*- coding: utf-8 -*-
# Copyright (C) Scott Coughlin (2017-2021)
#               Cardiff University (2021)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Extend :mod:`astropy.table` with the `GravitySpyTable`
"""

import re
from json import JSONDecodeError
from pathlib import Path
from urllib.parse import urlencode

import requests

from ..utils import mp as mp_utils
from .table import EventTable

__author__ = 'Scott Coughlin <scott.coughlin@ligo.org>'
__all__ = ['GravitySpyTable']

JSON_CONTENT_TYPE = "application/json"

# default gravity spy remote host
DEFAULT_HOST = "https://gravityspytools.ciera.northwestern.edu"

# API paths
SEARCH_PATH = "/search"
SIMILARITY_SEARCH_PATH = "/similarity_search_restful_API"

# ordered list of image durations supported by gravityspy.
# NOTE: the index of each element here is the map between
#       the 'urlX' column in the GravitySpyTable and the
#       duration of the relevant image, for some reason.
DURATIONS = (
    0.5,
    1.0,
    2.0,
    4.0,
)

# list of ERAs supported by similarity search API
ERA = {
    "ALL": (1126400000, 1584057618),
    "O1": (1126400000, 1137250000),
    "ER10": (1161907217, 1164499217),
    "O2a": (1164499217, 1219276818),
    "ER13": (1228838418, 1229176818),
    "ER14": (1235750418, 1238112018),
    "O3": (1238166018, 1272326418),
    "O3a": (1238166018, 1254009618),
    "O3b": (1256655642, 1272326418),
}

# regex to filter columns containing URLs
URL_COLUMN = re.compile(r"\Aurl[0-9]+\Z")


def _image_download_target(
    row,
    outdir,
    training_set=False,
    labelled_samples=False,
):
    """Construct where to download a GravitySpy image

    This returns a `str.format`-style template that just needs
    the duration to form a complete target path.
    """
    ifo = row["ifo"]
    id_ = row["gravityspy_id"]
    label = row["ml_label"] if training_set else ""
    stype = row["sample_type"] if labelled_samples else ""
    return str(
        outdir
        / label
        / stype
        / "{}_{}_spectrogram_{{}}.png".format(ifo, id_)
    )


def _download_image(bundle):
    url, target = bundle
    return target.write_bytes(requests.get(url).content)


class GravitySpyTable(EventTable):
    """A container for a table of Gravity Spy Events (as well as
    Events from the O1 Glitch Classification Paper whcih includes

    - PCAT
    - PC-LIB
    - WDF
    - WDNN
    - Karoo GP

    See also
    --------
    astropy.table.Table
        for details on parameters for creating a `GravitySpyTable`
    """

    # -- i/o ------------------------------------

    def download(
        self,
        download_path="download",
        nproc=1,
        download_durs=DURATIONS,
        training_set=False,
        labelled_samples=False,
    ):
        """Download image files associated with entries in a `GravitySpyTable`.

        Parameters
        ----------
        nproc : `int`, optional, default: 1
            number of CPUs to use for parallel file reading

        download_path : `str` optional, default: 'download'
            Specify where the images end up.

        download_durs : `list` optional, default: [0.5, 1.0, 2.0, 4.0]
            Specify exactly which durations you want to download
            default is to download all the avaialble GSpy durations.

        training_set : `bool`, optional
            if `True` download training set data

        labelled_samples : `bool`, optional
            if `True` download only labelled samples
        """
        download_path = Path(download_path)

        # work out which columns contain URLs
        url_columns = list(filter(URL_COLUMN.match, self.colnames))

        # construct a download target location for each requested image
        urls = dict()
        for row in self:
            if row["url1"] in {None, "", "?"}:  # skip rows without URLs
                continue

            # construct the abstract download target
            _download_target = _image_download_target(
                row,
                download_path,
                training_set=training_set,
                labelled_samples=labelled_samples,
            )

            # and map the remote URL to a concrete target
            # for each duration requested
            for col in url_columns:
                idx = int(col[3:])  # urlXXXX number
                duration = DURATIONS[idx-1]
                if duration in download_durs:  # image requested
                    urls[row[col]] = Path(_download_target.format(duration))

        # create all of the directories needed up front
        for path in urls.values():
            path.parent.mkdir(exist_ok=True, parents=True)

        # download the images
        mp_utils.multiprocess_with_queues(
            nproc,
            _download_image,
            urls.items(),
        )

        # and return the list of new files we have
        return sorted(map(str, urls.values()))

    @classmethod
    def search(
        cls,
        gravityspy_id,
        howmany=10,
        era="ALL",
        ifos=("H1", "L1"),
        database="similarity_index_o3",
        host=DEFAULT_HOST,
        **kwargs,
    ):
        """Perform a GravitySpy 'Similarity Search' for the given ID.

        Parameters
        ----------
        gravityspy_id : `str`
            the unique 10 character hash that identifies a Gravity Spy image

        howmany : `int`, optional
            number of similar images you would like

        era : `str`, optional
            which era to search, see online for more information

        ifos : `tuple`, str`, optional
            the list of interferometers to include in the search

        database : `str`, optional
            the database to query

        host : `str`, optional
            the URL (scheme and FQDN) of the Gravity Spy host to query

        **kwargs
            all other kwargs are passed to `requests.get` to perform
            the actual remote communication.

        Returns
        -------
        table : `GravitySpyTable`
            a `GravitySpyTable` containing similar events based on
            an evaluation of the Euclidean distance of the input image
            to all other images in some Feature Space

        Notes
        -----
        For an online version, and documentation of the search, see

        https://gravityspytools.ciera.northwestern.edu/search/
        """
        base = host + SEARCH_PATH + SIMILARITY_SEARCH_PATH

        if isinstance(ifos, str):  # 'H1L1' -> ['H1', 'L1']
            ifos = [ifos[i:i+2] for i in range(0, len(ifos), 2)]

        query = urlencode({
            'howmany': howmany,
            'imageid': gravityspy_id,
            'era': "event_time BETWEEN {} AND {}".format(*ERA[era]),
            'ifo': ", ".join(map(repr, ifos)),
            'database': database,
        })
        url = '{}/?{}'.format(base, query)

        response = requests.get(url, **kwargs)
        response.raise_for_status()  # check the response
        try:
            return GravitySpyTable(response.json())
        except JSONDecodeError:  # that didn't work
            # if the response was actually HTML, something terrible happened
            if response.headers["Content-Type"] != JSON_CONTENT_TYPE:
                raise requests.HTTPError(
                    "response from {} was '200 OK' but the content is HTML, "
                    "please check the request parameters".format(
                        response.url,
                    ),
                    response=response,
                )
            raise



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014-2020)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.signal.window`
"""

import numpy
import pytest

from scipy.signal import get_window as scipy_get_window

from ...testing import utils
from .. import window

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


@pytest.mark.parametrize(("win", "Nx", "out"), [
    ("hann", 10, scipy_get_window("hann", 10)),
    (24, 10, scipy_get_window(("kaiser", 24), 10)),
    (("kaiser", 24), 10, scipy_get_window(("kaiser", 24), 10)),
    (range(10), 10, numpy.arange(10)),
])
def test_get_window(win, Nx, out):
    numpy.testing.assert_array_equal(
        window.get_window(win, Nx),
        out,
    )


@pytest.mark.parametrize(("win", "Nx", "exc"), [
    (range(11), 10, "window array wrong size"),
    ([[0, 1], [1, 2]], 4, "multi-dimensional windows are not supported"),
])
def test_get_window_error(win, Nx, exc):
    with pytest.raises(
        AssertionError,
        match=exc,
    ):
        window.get_window(win, Nx)


@pytest.mark.parametrize(("in_, out"), [
    ("Han", "hann"),
])
def test_canonical_name(in_, out):
    assert window.canonical_name(in_) == out


def test_canonical_name_error():
    with pytest.raises(
        ValueError,
        match="^no window function in scipy.signal equivalent to 'blah'$",
    ):
        window.canonical_name('blah')


@pytest.mark.parametrize(("name", "overlap"), [
    ("hann", 0.5),
    ("Hamming", 0.5),
])
def test_recommended_overlap(name, overlap):
    assert window.recommended_overlap(name) == overlap


def test_recommended_overlap_nfft():
    assert window.recommended_overlap('barthann', nfft=128) == 64


def test_recommended_overlap_error():
    with pytest.raises(
        ValueError,
        match="^no recommended overlap for 'kaiser' window$",
    ):
        window.recommended_overlap('kaiser')


def test_planck():
    series = numpy.ones(64)
    wseries = series * window.planck(64, nleft=5, nright=5)
    assert wseries[0] == 0
    assert wseries[-1] == 0
    utils.assert_allclose(wseries[5:59], series[5:59])
